# -*- coding: utf-8 -*-
"""
本模块功能：期权定价理论理论计算函数包
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2020年7月16日
最新修订日期：2020年8月5日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
作者邮件：wdehong2000@163.com
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""
#==============================================================================
#统一屏蔽一般性警告
import warnings; warnings.filterwarnings("ignore")    
from siat.common import *
from siat.translate import *
from siat.grafix import *
from siat.security_prices import *
#==============================================================================
import matplotlib.pyplot as plt

#处理绘图汉字乱码问题
import sys; czxt=sys.platform
if czxt in ['win32','win64']:
    plt.rcParams['font.sans-serif'] = ['SimHei']  # 设置默认字体
    mpfrc={'font.family': 'SimHei'}

if czxt in ['darwin']: #MacOSX
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family': 'Heiti TC'}

if czxt in ['linux']: #website Jupyter
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family':'Heiti TC'}

# 解决保存图像时'-'显示为方块的问题
plt.rcParams['axes.unicode_minus'] = False 
#==============================================================================
def bs_call(S0,X,Days,r0,sigma,printout=True):
    """
    功能：计算无红利支付的欧式期权B-S定价模型，看涨期权
    注意：
    S0：标的物资产的当前价格，X为其行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    """
    from scipy import log,exp,sqrt,stats
    #Days为距离到期日的日历日天数
    T=Days/365.
    r=log(r0+1)
    
    d1=(log(S0/X)+(r+sigma*sigma/2.)*T)/(sigma*sqrt(T))
    d2=d1-sigma*sqrt(T)
    
    C0=S0*stats.norm.cdf(d1)-X*exp(-r*T)*stats.norm.cdf(d2)
    
    if not printout: return C0
    print("\n===== Black-Scholes期权定价 =====")
    print("适用情形： 欧式期权，标的资产无红利收益")
    print("标的资产行权价:",X)
    print("标的资产现价  :",S0)
    print("标的资产价格的年化波动率:",round(sigma,4))    
    print("距离到期日的年数        :",round(T,2))
    print("连续计算的无风险利率    :",round(r,4)*100,'\b%')
    print("看涨期权的预期价格      :",round(C0,2))
    
    return C0
    
if __name__=='__main__':
    S0=40
    X=42
    Days=183
    r0=0.03
    sigma=0.02
    C0=bs_call(40,42,183,0.015,0.02)

#==============================================================================
def bsm_call(S0,X,Days,r0,sigma,Days1=0,div1=0,printout=True):
    """
    功能：计算有红利支付的欧式期权B-S定价模型，看涨期权
    注意：
    S0：标的物资产的当前价格，X为其行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    from scipy import log,exp
    #Days1为距离到期日的日历日天数
    T=Days/365.
    T1=Days1/365.
    r=log(r0+1)    
    #调整标的物当前价
    S=S0-exp(-r*T1)*div1
    
    #调用BS模型计算
    C=bs_call(S,X,Days,r0,sigma,printout=False)

    if not printout: return C
    print("\n=== Black-Scholes-Merton期权定价 ===")
    print("适用情形： 欧式期权，标的资产有红利收益")
    print("标的资产行权价:",X)
    print("标的资产现价  :",S0)
    print("标的资产价格的年化波动率:",round(sigma,4))    
    print("距离到期日的年数        :",round(T,2))
    print("连续计算的无风险利率    :",round(r,4)*100,'\b%')    
    print("红利及距离到期日的年数  :",div1,"@",round(T1,2))
    print("看涨期权的预期价格      :",round(C,2))
    
    return C
    
if __name__=='__main__':
    S0=42
    X=40
    Days=183
    r0=0.03
    sigma=0.02
    dv1=1.5
    Days1=183
    C=bsm_call(42,40,183,0.015,0.02,183,1.5)
    C0=bsm_call(42,40,183,0.015,0.23)

#==============================================================================
def bs_put(S0,X,Days,r0,sigma,printout=True):
    """
    功能：计算无红利支付的欧式期权B-S定价模型，看跌期权
    注意：
    S0：标的物资产的当前价格，X为其行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    """
    from scipy import log,exp,sqrt,stats
    #Days为距离到期日的日历日天数
    T=Days/365.
    r=log(r0+1)
    
    d1=(log(S0/X)+(r+sigma*sigma/2.)*T)/(sigma*sqrt(T))
    d2=d1-sigma*sqrt(T)
    
    P0=-S0*stats.norm.cdf(-d1)+X*exp(-r*T)*stats.norm.cdf(-d2)

    if not printout: return P0
    print("\n===== Black-Scholes期权定价 =====")
    print("适用情形： 欧式期权，标的资产无红利收益")
    print("标的资产行权价:",X)
    print("标的资产现价  :",S0)
    print("标的资产价格的年化波动率:",round(sigma,4))    
    print("距离到期日的年数        :",round(T,2))
    print("连续计算的无风险利率    :",round(r,4)*100,'\b%')
    print("看跌期权的预期价格      :",round(P0,2))
    
    return P0
    
if __name__=='__main__':
    S0=40
    X=42
    Days=183
    r0=0.03
    sigma=0.02
    P0=bs_put(40,42,183,0.015,0.02)    

#==============================================================================
def bsm_put(S0,X,Days,r0,sigma,Days1=0,div1=0,printout=True):
    """
    功能：计算有红利支付的欧式期权B-S定价模型，看跌期权
    注意：
    S0：标的物资产的当前价格，X为其行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    from scipy import log,exp,sqrt
    #Days为距离到期日的日历日天数
    T=Days/365.
    T1=Days1/365.
    r=log(r0+1)
    
    S=S0-exp(-r*T1)*div1
    
    #调用BS模型计算
    P=bs_put(S,X,Days,r0,sigma,printout=False)

    if not printout: return P
    print("\n=== Black-Scholes-Merton期权定价 ===")
    print("适用情形： 欧式期权，标的资产有红利收益")
    print("标的资产行权价:",X)
    print("标的资产现价  :",S0)
    print("标的资产价格的年化波动率:",round(sigma,4))    
    print("距离到期日的年数        :",round(T,2))
    print("连续计算的无风险利率    :",round(r,4)*100,'\b%')
    print("红利及距离到期日的年数  :",div1,"@",round(T1,2))
    print("看跌期权的预期价格      :",round(P,2))
    
    return P
    
if __name__=='__main__':
    S0=42
    X=40
    Days=183
    r0=0.03
    sigma=0.02
    P=bsm_put(42,40,183,0.015,0.23,90,1.5)   
    P0=bsm_put(42,40,183,0.015,0.23)

#==============================================================================
def bsm_put_aprice(Srange,X,Days,r0,sigma,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看跌期权，当前价格为变化范围
    注意：
    Srange：标的物资产的当前价格范围，默认20等分后作为横轴绘图
    X：期权的行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=Srange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_put_aprice): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_put_aprice): not enough range for target,",trange)
        return        
    
    #确定起始位置和间隔大小
    tstart=trange[0]; tend=trange[1]
    if len(trange) >=3: tstep=trange[2]    
    else: tstep=(tend-tstart)/20.
    #横轴点列表
    import numpy as np
    tlist=np.arange(tstart,tend+tstep,tstep)

    #循环计算各点数值
    import pandas as pd
    df=pd.DataFrame(columns=['Option Price','Asset Price','Strike Price', \
                             'Days to Maturity','Annual RF', \
                             'Annual Sigma','Div to Maturity','Dividend'])
    for t in tlist:
        #通用修改点
        op=bsm_put(t,X,Days,r0,sigma,Days1,div1,printout=False)
        s=pd.Series({'Option Price':op,'Asset Price':t,'Strike Price':X, \
                     'Days to Maturity':Days,'Annual RF':r0, \
                     'Annual Sigma':sigma,'Div to Maturity':Days1, \
                     'Dividend':div1})
        df=df.append(s,ignore_index=True)        
    #通用修改点
    df2=df.set_index(['Asset Price'])  
    if not graph: return df2      
    
    #绘图
    #通用修改点
    colname='Option Price'; collabel='看跌期权'
    ylabeltxt='期权价格'
    titletxt='期权价格的变化趋势图'
    #通用修改点
    fn1='标的物市场价格'
    fn2='\n[期权信息]行权价='+str(X)+', 距离到期天数='+str(Days)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 年化波动率%='+str(round(sigma*100.,2))
    if div1==0:
        fn4='\n本产品无红利收益'
    else:
        fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plot_line(df2,colname,collabel,ylabeltxt,titletxt,footnote)
    
    return df2
    
if __name__=='__main__':
    Srange=[30,50,1]
    X=40
    Days=183
    r0=0.03
    sigma=0.02
    Days1=0
    div1=0
    pdf=bsm_put_aprice(Srange,40,183,0.015,0.23,90,1.5)    

#==============================================================================
def bsm_call_aprice(Srange,X,Days,r0,sigma,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看涨期权，当前价格为变化范围
    注意：
    Srange：标的物资产的当前价格范围，默认20等分后作为横轴绘图
    X：期权的行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=Srange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_call_aprice): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_call_aprice): not enough range for target,",trange)
        return        
    
    #确定起始位置和间隔大小
    tstart=trange[0]; tend=trange[1]
    if len(trange) >=3: tstep=trange[2]    
    else: tstep=(tend-tstart)/20.
    #横轴点列表
    import numpy as np
    tlist=np.arange(tstart,tend+tstep,tstep)

    #循环计算各点数值
    import pandas as pd
    df=pd.DataFrame(columns=['Option Price','Asset Price','Strike Price', \
                             'Days to Maturity','Annual RF', \
                             'Annual Sigma','Div to Maturity','Dividend'])
    for t in tlist:
        #通用修改点
        op=bsm_call(t,X,Days,r0,sigma,Days1,div1,printout=False)
        s=pd.Series({'Option Price':op,'Asset Price':t,'Strike Price':X, \
                     'Days to Maturity':Days,'Annual RF':r0, \
                     'Annual Sigma':sigma,'Div to Maturity':Days1, \
                     'Dividend':div1})
        df=df.append(s,ignore_index=True)        
    #通用修改点
    df2=df.set_index(['Asset Price'])  
    if not graph: return df2      
    
    #绘图
    #通用修改点
    colname='Option Price'; collabel='看涨期权'
    ylabeltxt='期权价格'
    titletxt='期权价格的变化趋势图'
    #通用修改点
    fn1='标的物市场价格'
    fn2='\n[期权信息]行权价='+str(X)+', 距离到期天数='+str(Days)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 年化波动率%='+str(round(sigma*100.,2))
    if div1==0:
        fn4='\n本产品无红利收益'
    else:
        fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plot_line(df2,colname,collabel,ylabeltxt,titletxt,footnote)
    
    return df2
    
if __name__=='__main__':
    Srange=[30,50,1]
    X=40
    Days=183
    r0=0.03
    sigma=0.02
    Days1=0
    div1=0
    cdf=bsm_call_aprice(Srange,40,183,0.015,0.23,90,1.5)    

#==============================================================================
def bsm_aprice(Srange,X,Days,r0,sigma,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看涨/看跌期权，当前价格为变化范围
    注意：
    Srange：标的物资产的当前价格范围，默认20等分后作为横轴绘图
    X：期权的行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=Srange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_aprice): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_aprice): not enough range for target,",trange)
        return        
    
    #看涨期权
    df1=bsm_call_aprice(Srange,X,Days,r0,sigma,Days1=0,div1=0,graph=False)
    #看跌期权
    df2=bsm_put_aprice(Srange,X,Days,r0,sigma,Days1=0,div1=0,graph=False)     
    
    #绘图
    #通用修改点
    ticker1='看涨期权'; colname1='Option Price'; label1='期权-C-'+str(X)
    ticker2='看跌期权'; colname2='Option Price'; label2='期权-P-'+str(X)
    ylabeltxt='期权价格'
    titletxt='期权价格的变化趋势图'
    #通用修改点
    fn1='标的物市场价格-->'
    fn2='\n[期权信息]行权价='+str(X)+', 距离到期天数='+str(Days)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 年化波动率%='+str(round(sigma*100.,2))
    if div1==0:
        fn4='\n本产品无红利收益'
    else:
        fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plot_line2_coaxial(df1,ticker1,colname1,label1, \
                       df2,ticker2,colname2,label2, \
                    ylabeltxt,titletxt,footnote)
    return
    
if __name__=='__main__':
    Srange=[30,50,1]
    X=40
    Days=183
    r0=0.03
    sigma=0.02
    Days1=0
    div1=0
    bsm_aprice(Srange,40,183,0.015,0.23,90,1.5)  
    bsm_aprice([30,50],40,183,0.015,0.23,90,1.5)  

#==============================================================================
def bsm_put_maturity(S0,X,Dayrange,r0,sigma,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看跌期权，距离到期日天数为变化范围
    注意：
    S0：标的物资产的当前价格
    X：期权的行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Dayrange：距离到期日的天数范围，默认变化间隔为20分之一取整
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=Dayrange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_put_maturity): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_put_maturity): not enough range for target,",trange)
        return        
    
    #确定起始位置和间隔大小
    tstart=int(trange[0]); tend=int(trange[1])
    if len(trange) >=3: tstep=trange[2]    
    else: tstep=int((tend-tstart)/20)
    #横轴点列表
    #import numpy as np
    tlist=range(tstart,tend+tstep,tstep)

    #循环计算各点数值
    import pandas as pd
    df=pd.DataFrame(columns=['Option Price','Asset Price','Strike Price', \
                             'Days to Maturity','Annual RF', \
                             'Annual Sigma','Div to Maturity','Dividend'])
    for t in tlist:
        #通用修改点
        op=bsm_put(S0,X,t,r0,sigma,Days1,div1,printout=False)
        s=pd.Series({'Option Price':op,'Asset Price':S0,'Strike Price':X, \
                     'Days to Maturity':t,'Annual RF':r0, \
                     'Annual Sigma':sigma,'Div to Maturity':Days1, \
                     'Dividend':div1})
        df=df.append(s,ignore_index=True)        
    #通用修改点
    df2=df.set_index(['Days to Maturity'])  
    if not graph: return df2      
    
    #绘图
    #通用修改点
    colname='Option Price'; collabel='看跌期权'
    ylabeltxt='期权价格'
    titletxt='期权价格的变化趋势图'
    #通用修改点
    fn1='<--距离到期日的天数'
    fn2='\n[期权信息]行权价='+str(X)+', 标的物市价='+str(S0)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 年化波动率%='+str(round(sigma*100.,2))
    if div1==0:
        fn4='\n本产品无红利收益'
    else:
        fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plot_line(df2,colname,collabel,ylabeltxt,titletxt,footnote)
    
    return df2
    
if __name__=='__main__':
    S0=42
    X=40
    Dayrange=[200,50]
    r0=0.015
    sigma=0.23
    Days1=90
    div1=1.5
    pdf=bsm_put_maturity(42,40,[200,50],0.015,0.23,90,1.5)    

#==============================================================================
def bsm_call_maturity(S0,X,Dayrange,r0,sigma,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看涨期权，距离到期日天数为变化范围
    注意：
    S0：标的物资产的当前价格
    X：期权的行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Dayrange：距离到期日的天数范围，默认变化间隔为20分之一取整
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=Dayrange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_call_maturity): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_call_maturity): not enough range for target,",trange)
        return        
    
    #确定起始位置和间隔大小
    tstart=int(trange[0]); tend=int(trange[1])
    if len(trange) >=3: tstep=trange[2]    
    else: tstep=int((tend-tstart)/20)
    #横轴点列表
    #import numpy as np
    tlist=range(tstart,tend+tstep,tstep)

    #循环计算各点数值
    import pandas as pd
    df=pd.DataFrame(columns=['Option Price','Asset Price','Strike Price', \
                             'Days to Maturity','Annual RF', \
                             'Annual Sigma','Div to Maturity','Dividend'])
    for t in tlist:
        #通用修改点
        op=bsm_call(S0,X,t,r0,sigma,Days1,div1,printout=False)
        s=pd.Series({'Option Price':op,'Asset Price':S0,'Strike Price':X, \
                     'Days to Maturity':t,'Annual RF':r0, \
                     'Annual Sigma':sigma,'Div to Maturity':Days1, \
                     'Dividend':div1})
        df=df.append(s,ignore_index=True)        
    #通用修改点
    df2=df.set_index(['Days to Maturity'])  
    if not graph: return df2      
    
    #绘图
    #通用修改点
    colname='Option Price'; collabel='看涨期权'
    ylabeltxt='期权价格'
    titletxt='期权价格的变化趋势图'
    #通用修改点
    fn1='<--距离到期日的天数'
    fn2='\n[期权信息]行权价='+str(X)+', 标的物市价='+str(S0)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 年化波动率%='+str(round(sigma*100.,2))
    if div1==0:
        fn4='\n本产品无红利收益'
    else:
        fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plot_line(df2,colname,collabel,ylabeltxt,titletxt,footnote)
    
    return df2
    
if __name__=='__main__':
    S0=42
    X=40
    Dayrange=[200,50]
    r0=0.015
    sigma=0.23
    Days1=90
    div1=1.5
    cdf=bsm_call_maturity(42,40,[200,50],0.015,0.23,90,1.5)    

#==============================================================================
if __name__=='__main__':
    S0=42
    X=40
    Dayrange=[200,50]
    r0=0.015
    sigma=0.23
    Days1=90
    div1=1.5

def bsm_maturity(S0,X,Dayrange,r0,sigma,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看涨/看跌期权，距离到期日天数为变化范围
    注意：
    S0：标的物资产的当前价格
    X：期权的行权价
    sigma：标的物资产价格的年化标准差
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Dayrange：距离到期日的天数范围，默认间隔为20分之一取证
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=Dayrange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_maturity): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_maturity): not enough range for target,",trange)
        return        
    
    #看涨期权
    df1=bsm_call_maturity(S0,X,Dayrange,r0,sigma,Days1=0,div1=0,graph=False)
    df1.sort_index(ascending=False, inplace=True)
    #看跌期权
    df2=bsm_put_maturity(S0,X,Dayrange,r0,sigma,Days1=0,div1=0,graph=False)
    df2.sort_index(ascending=False, inplace=True)  
    #合并
    import pandas as pd
    df=pd.merge(df1,df2,how='inner',left_index=True,right_index=True,sort=True)
    #df['到期日']=df.index
    #df.sort_values('到期日',ascending=False,inplace=True)
    
    #绘图    
    plt.title('期权价格的变化趋势图')
    plt.ylabel('期权价格')
    
    df['Option Price_x'].plot(label='看涨期权',ls='-')
    df['Option Price_y'].plot(label='看跌期权',ls='-.')
    
    fn1='距离到期日的天数-->'
    fn2='\n[期权信息]行权价='+str(X)+', 标的物市价='+str(S0)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 年化波动率%='+str(round(sigma*100.,2))
    if div1==0: fn4='\n本产品无红利收益'
    else: fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plt.xlabel(footnote)  
    
    #让横轴逆序从大到小显示，正常顺序为从小到大
    plt.gca().invert_xaxis()
    plt.legend()
    plt.show()    
    
    return
    
if __name__=='__main__':
    S0=42
    X=40
    Dayrange=[200,50]
    r0=0.015
    sigma=0.23
    Days1=90
    div1=1.5
    bsm_maturity(S0,40,[200,5],0.015,0.23,90,1.5)  
    bsm_maturity(42,40,[50,200],0.015,0.23,90,1.5)  

#==============================================================================
def bsm_put_sigma(S0,X,Days,r0,sigmarange,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看跌期权，年化波动率为变化范围
    注意：
    S0：标的物资产的当前价格
    X：期权的行权价
    sigmarange：标的物资产价格的年化标准差范围，默认为区间的20分之一作为间隔
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=sigmarange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_put_sigma): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_put_sigma): not enough range for target,",trange)
        return        
    
    #确定起始位置和间隔大小
    tstart=trange[0]; tend=trange[1]
    if len(trange) >=3: tstep=trange[2]    
    else: tstep=(tend-tstart)/20.
    #横轴点列表
    import numpy as np
    tlist=np.arange(tstart,tend+tstep,tstep)

    #循环计算各点数值
    import pandas as pd
    df=pd.DataFrame(columns=['Option Price','Asset Price','Strike Price', \
                             'Days to Maturity','Annual RF', \
                             'Annual Sigma','Div to Maturity','Dividend'])
    for t in tlist:
        #通用修改点
        op=bsm_put(S0,X,Days,r0,t,Days1,div1,printout=False)
        s=pd.Series({'Option Price':op,'Asset Price':S0,'Strike Price':X, \
                     'Days to Maturity':Days,'Annual RF':r0, \
                     'Annual Sigma':t,'Div to Maturity':Days1, \
                     'Dividend':div1})
        df=df.append(s,ignore_index=True)        
    #通用修改点
    df2=df.set_index(['Annual Sigma'])  
    if not graph: return df2      
    
    #绘图
    #通用修改点
    colname='Option Price'; collabel='看跌期权'
    ylabeltxt='期权价格'
    titletxt='期权价格的变化趋势图'
    #通用修改点
    fn1='波动率-->'
    fn2='\n[期权信息]行权价='+str(X)+', 标的物市价='+str(S0)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 距离到期日天数='+str(Days)
    if div1==0:
        fn4='\n本产品无红利收益'
    else:
        fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plot_line(df2,colname,collabel,ylabeltxt,titletxt,footnote)
    
    return df2
    
if __name__=='__main__':
    S0=42
    X=40
    Days=183
    r0=0.015
    sigmarange=[0.1,0.4]
    Days1=90
    div1=1.5
    pdf=bsm_put_sigma(42,40,183,0.015,[0.1,0.4],90,1.5)    


#==============================================================================
def bsm_call_sigma(S0,X,Days,r0,sigmarange,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看涨期权，年化波动率为变化范围
    注意：
    S0：标的物资产的当前价格
    X：期权的行权价
    sigmarange：标的物资产价格的年化标准差范围，默认为区间的20分之一作为间隔
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=sigmarange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_call_sigma): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_call_sigma): not enough range for target,",trange)
        return        
    
    #确定起始位置和间隔大小
    tstart=trange[0]; tend=trange[1]
    if len(trange) >=3: tstep=trange[2]    
    else: tstep=(tend-tstart)/20.
    #横轴点列表
    import numpy as np
    tlist=np.arange(tstart,tend+tstep,tstep)

    #循环计算各点数值
    import pandas as pd
    df=pd.DataFrame(columns=['Option Price','Asset Price','Strike Price', \
                             'Days to Maturity','Annual RF', \
                             'Annual Sigma','Div to Maturity','Dividend'])
    for t in tlist:
        #通用修改点
        op=bsm_call(S0,X,Days,r0,t,Days1,div1,printout=False)
        s=pd.Series({'Option Price':op,'Asset Price':S0,'Strike Price':X, \
                     'Days to Maturity':Days,'Annual RF':r0, \
                     'Annual Sigma':t,'Div to Maturity':Days1, \
                     'Dividend':div1})
        df=df.append(s,ignore_index=True)        
    #通用修改点
    df2=df.set_index(['Annual Sigma'])  
    if not graph: return df2      
    
    #绘图
    #通用修改点
    colname='Option Price'; collabel='看涨期权'
    ylabeltxt='期权价格'
    titletxt='期权价格的变化趋势图'
    #通用修改点
    fn1='波动率-->'
    fn2='\n[期权信息]行权价='+str(X)+', 标的物市价='+str(S0)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 距离到期日天数='+str(Days)
    if div1==0:
        fn4='\n本产品无红利收益'
    else:
        fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    plot_line(df2,colname,collabel,ylabeltxt,titletxt,footnote)
    
    return df2
    
if __name__=='__main__':
    S0=42
    X=40
    Days=183
    r0=0.015
    sigmarange=[0.1,0.4]
    Days1=90
    div1=1.5
    cdf=bsm_call_sigma(42,40,183,0.015,[0.1,0.4],90,1.5)  

#==============================================================================
def bsm_sigma(S0,X,Days,r0,sigmarange,Days1=0,div1=0,graph=True):
    """
    功能：计算有红利支付的欧式期权BSM定价模型，看涨/看跌期权，波动率为变化范围
    注意：
    S0：标的物资产的当前价格
    X：期权的行权价
    sigmarange：标的物资产价格的年化标准差范围
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    Days：距离到期日的天数
    Days1：红利发放时距离到期日的天数，需要转换为年数
    div1：红利金额
    """
    #通用修改点
    trange=sigmarange
    #检查是否为列表
    if not isinstance(trange,list):
        print("#Error(bsm_sigma): target is not a range,",trange)
        return
    if len(trange) < 2:
        print("#Error(bsm_sigma): not enough range for target,",trange)
        return        
    
    #看涨期权
    df1=bsm_call_sigma(S0,X,Days,r0,sigmarange,Days1=0,div1=0,graph=False)
    df1.sort_index(ascending=True, inplace=True)
    #看跌期权
    df2=bsm_put_sigma(S0,X,Days,r0,sigmarange,Days1=0,div1=0,graph=False)
    df2.sort_index(ascending=True, inplace=True)  
    
    #绘制双线图
    titletxt='期权价格的变化趋势图'
    colname1='Option Price'; label1='看涨期权'
    colname2='Option Price'; label2='看跌期权'
    ylabeltxt='期权价格'
    
    fn1='波动率-->'
    fn2='\n[期权信息]行权价='+str(X)+', 标的物市价='+str(S0)
    fn3='\n年化无风险利率%='+str(round(r0*100.,2))+', 距离到期日天数='+str(Days)
    if div1==0: fn4='\n本产品无红利收益'
    else: fn4='\n发放红利时间距离到期天数='+str(Days1)+', 红利='+str(div1)
    footnote=fn1+fn2+fn3+fn4
    
    plot_2lines(df1,colname1,label1,df2,colname2,label2, \
                ylabeltxt,titletxt,footnote)
    
    return
    
if __name__=='__main__':
    S0=42
    X=40
    Days=183
    r0=0.015
    sigmarange=[0.1,0.4]
    Days1=90
    div1=1.5
    bsm_sigma(42,40,183,0.015,[0.1,0.4],90,1.5)  

#==============================================================================
def iv_call_bsm(aop,S0,X,Days,r0,Days1=0,div1=0,precision=0.01,printout=True):
    """
    功能：基于BSM模型，二分迭代法，计算隐含波动率，看涨期权
    aop：实际期权价格
    S0：标的物当前市价
    X：标的物行权价
    Days：距离到期日的天数
    r0：年华无风险利率
    Days1：预期红利收益发放日期距离到期日的天数，默认=0
    div1：预期红利收益金额，默认=0
    printout：是否显示计算结果，默认=True
    """
    k=1
    volLow=0.001    #设置波动率的最低值
    volHigh=1.0     #设置波动率的最高值
    
    #波动率最低值对应的期权价格
    #cLow=bsCall(S,X,T,r,volLow)
    cLow=bsm_call(S0,X,Days,r0,volLow,Days1,div1,printout=False)
    
    #波动率最高值对应的期权价格
    #cHigh=bsCall(S,X,T,r,volHigh)  
    cHigh=bsm_call(S0,X,Days,r0,volHigh,Days1,div1,printout=False)
    
    #防止出现死循环
    if cLow > aop or cHigh < aop: 
        print("#Error(iv_call_bsm): Option price not reasonable,",aop)
        return None        
        #raise ValueError    
    
    while k ==1:
        #cLow=bsCall(S,X,T,r,volLow)
        cLow=bsm_call(S0,X,Days,r0,volLow,Days1,div1,printout=False)
        #cHigh=bsCall(S,X,T,r,volHigh)
        cHigh=bsm_call(S0,X,Days,r0,volHigh,Days1,div1,printout=False)
        
        #取波动率的高低均值
        volMid=(volLow+volHigh)/2.0
        #cMid=bsCall(S,X,T,r,volMid) 
        cMid=bsm_call(S0,X,Days,r0,volMid,Days1,div1,printout=False)
        
        #满足期权价格误差精度要求（0.01以下）则结束循环
        if abs(cHigh-cLow) < precision: k=2
        #否则，缩小范围，继续循环
        elif cMid>aop: volHigh=volMid
        else: volLow=volMid        
    
    iv=round(volMid,4)
    if not printout: return iv
    
    #显示
    print("\n=== 隐含波动率: 二叉树迭代 ===")
    print("看涨期权:")
    print("  期权现价        :",aop)
    print("  标的资产现价    :",S0)
    print("  标的资产行权价  :",X)
    print("  距离到期日的天数:",Days)
    if not (div1 == 0):
        print("  预期红利                :",div1)
        print("  红利发放距离到期日的天数:",Days1)
    print("隐含波动率:")
    print("  预计的年化波动率  :",iv)
    print("  对应的期权价格范围:",round(cLow,3),'-',round(cHigh,3))
    print("  迭代精度          :",precision)
    return iv

#==============================================================================
def iv_put_bsm(aop,S0,X,Days,r0,Days1=0,div1=0,precision=0.01,printout=True):
    """
    功能：基于BSM模型，二分迭代法，计算隐含波动率，看跌期权
    aop：实际期权价格
    S0：标的物当前市价
    X：标的物行权价
    Days：距离到期日的天数
    r0：年华无风险利率
    Days1：预期红利收益发放日期距离到期日的天数，默认=0
    div1：预期红利收益金额，默认=0
    printout：是否显示计算结果，默认=True
    """
    k=1
    volLow=0.001    #设置波动率的最低值
    volHigh=1.0     #设置波动率的最高值
    
    #波动率最低值对应的期权价格
    #cLow=bsCall(S,X,T,r,volLow)
    cLow=bsm_put(S0,X,Days,r0,volLow,Days1,div1,printout=False)
    
    #波动率最高值对应的期权价格
    #cHigh=bsCall(S,X,T,r,volHigh)  
    cHigh=bsm_put(S0,X,Days,r0,volHigh,Days1,div1,printout=False)
    
    #防止出现死循环
    if cLow > aop or cHigh < aop: 
        print("#Error(iv_put_bsm): Option price not reasonable,",aop)
        return None
        #raise ValueError    
    
    while k ==1:
        #cLow=bsCall(S,X,T,r,volLow)
        cLow=bsm_put(S0,X,Days,r0,volLow,Days1,div1,printout=False)
        #cHigh=bsCall(S,X,T,r,volHigh)
        cHigh=bsm_put(S0,X,Days,r0,volHigh,Days1,div1,printout=False)
        
        #取波动率的高低均值
        volMid=(volLow+volHigh)/2.0
        #cMid=bsCall(S,X,T,r,volMid) 
        cMid=bsm_put(S0,X,Days,r0,volMid,Days1,div1,printout=False)
        
        #满足期权价格误差精度要求（precision）则结束循环
        if abs(cHigh-cLow) < precision: k=2
        #否则，缩小范围，继续循环
        elif cMid>aop: volHigh=volMid
        else: volLow=volMid        
    
    iv=round(volMid,4)
    if not printout: return iv
    
    #显示
    print("\n=== 隐含波动率: 二叉树迭代 ===")
    print("看跌期权:")
    print("  期权现价        :",aop)
    print("  标的资产现价    :",S0)
    print("  标的资产行权价  :",X)
    print("  距离到期日的天数:",Days)
    if not (div1 == 0):
        print("  预期红利                :",div1)
        print("  红利发放距离到期日的天数:",Days1)
    print("隐含波动率:")
    print("  预计的年化波动率  :",iv)
    print("  对应的期权价格范围:",round(cLow,3),'-',round(cHigh,3))
    print("  迭代精度          :",precision)

    return iv

#==============================================================================
def binomial_american_call(S0,X,Days,r0,sigma,q0=0,steps=200,printout=True):
    """
    功能：计算有红利支付的美式期权二叉树定价模型，看涨期权
    注意：
    S0：标的物资产的当前价格
    X为其行权价
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    sigma：标的物资产价格的年化标准差
    q0：年化红利收益率，由于美式期权可能提前行权，故不考虑发放日期
    steps：二叉树的步骤
    """
    from scipy import log,exp
    #Days1为距离到期日的日历日天数
    t=Days/365.
    r=log(r0+1)  
    q=log(q0+1)
    #调整标的物当前价
    S=S0    

    import numpy as np
    u=np.exp(sigma*np.sqrt(t/steps)); d=1/u
	
    P=(np.exp((r-q)*t/steps)-d)/(u-d)
    prices=np.zeros(steps+1)
    c_values=np.zeros(steps+1)
    prices[0]=S*d**steps
    c_values[0]=np.maximum(prices[0]-X,0)
	
    for i in range(1,steps+1):
        prices[i]=prices[i-1]*(u**2)
        c_values[i]=np.maximum(prices[i]-X,0)
	
    for j in range(steps,0,-1):
        for i in range(0,j):
            prices[i]=prices[i+1]*d
            #c_values[i]=np.maximum((P*c_values[i+1]+(1-P)*c_values[i])/np.exp(r*t/steps),prices[i]-X)
            c_values[i]=np.maximum((P*c_values[i+1]+(1-P)*c_values[i])/np.exp((r-q)*t/steps),prices[i]-X)
    C=round(c_values[0],2)

    if not printout: return C
    print("\n===== 二叉树期权定价 =====")
    print("适用情形： 美式期权，标的资产有红利")
    print("标的资产行权价        :",X)
    print("标的资产现价          :",S0)
    print("标的资产年化价格波动率:",round(sigma,4))    
    print("距离到期日的年数      :",round(t,2))
    print("连续计算的无风险收益率:",round(r,4)*100,'\b%')
    print("连续计算的红利率      :",round(q,4)*100,'\b%')
    print("二叉树迭代步数        :",steps)
    print("看涨期权的预期价格    :",round(C,2))    
    
    return C

#==============================================================================
def binomial_american_put(S0,X,Days,r0,sigma,q0=0,steps=200,printout=True):
    """
    功能：计算有红利支付的美式期权二叉树定价模型，看跌期权
    注意：
    S0：标的物资产的当前价格
    X为其行权价
    Days：距离到期日的天数，需要转换为距离到期日的年数=距离到期日的天数/365
    r0：年化无风险利率（需要转化为连续计算的无风险利率）
    sigma：标的物资产价格的年化标准差
    q0：年化红利收益率，由于美式期权可能提前行权，故不考虑发放日期
    steps：二叉树的步骤
    """
    from scipy import log,exp
    #Days1为距离到期日的日历日天数
    t=Days/365.
    r=log(r0+1)  
    q=log(q0+1)
    #调整标的物当前价
    S=S0   
    
    import numpy as np
    u=np.exp(sigma*np.sqrt(t/steps))
    d=1/u
    P=(np.exp((r-q)*t/steps)-d)/(u-d)
    prices=np.zeros(steps+1)
    c_values=np.zeros(steps+1)
    prices[0]=S*d**steps
    c_values[0]=np.maximum(X-prices[0],0)

    for i in range(1,steps+1):
        prices[i]=prices[i-1]*(u**2)
        c_values[i]=np.maximum(X-prices[i],0)

    for j in range(steps,0,-1):
        for i in range(0,j):
            prices[i]=prices[i+1]*d
            #c_values[i]=np.maximum((P*c_values[i+1]+(1-P)*c_values[i])/np.exp(r*t/steps),X-prices[i])#检查是否提前行权
            c_values[i]=np.maximum((P*c_values[i+1]+(1-P)*c_values[i])/np.exp((r-q)*t/steps),X-prices[i])#检查是否提前行权
    C=round(c_values[0],2)

    if not printout: return C
    print("\n===== 二叉树期权定价 =====")
    print("适用情形： 美式期权，标的资产有红利")
    print("标的资产行权价        :",X)
    print("标的资产现价          :",S0)
    print("标的资产年化价格波动率:",round(sigma,4))    
    print("距离到期日的年数      :",round(t,2))
    print("连续计算的无风险收益率:",round(r,4)*100,'\b%')
    print("连续计算的红利率      :",round(q,4)*100,'\b%')
    print("二叉树迭代步数        :",steps)
    print("看跌期权的预期价格    :",round(C,2)) 

    return C

#==============================================================================
#==============================================================================
#==============================================================================
#==============================================================================
if __name__=='__main__':
    ticker="AAPL"

def option_maturity(ticker,printout=True):
    """
    功能：获得期权的各个到期日期
    """
    import yfinance as yf
    opt = yf.Ticker(ticker)
    
    #获得期权的各个到期日
    try:
        exp_dates=opt.options
    except:
        print("#Error(option_maturity): failed to get maturity dates for option",ticker)
        return None
    datelist=list(exp_dates)
    if not printout:
        return datelist
    
    #显示结果
    print("\n===== 期权的时间序列 =====")
    print("标的资产:",ticker)
    print("到期日期:")
    
    num=len(datelist)
    for d in datelist:
        print(d,end='  ')
        pos=datelist.index(d)+1
        if (pos % 4 ==0) or (pos==num): print(' ')
    
    print("总计:",num,"个日期")
    import datetime
    today = datetime.date.today()    
    print("*** 数据来源: 雅虎财经,",today)
    
    return datelist

if __name__=='__main__':
    ticker='AAPL'
    datelist=option_maturity(ticker)
    datelist=option_maturity('AAPL')
    datelist=option_maturity('000001.SS')

#================================================================
if __name__=='__main__':
    ticker="AAPL"
    mdate="2020-11-06"

def option_chain(ticker,mdate,printout=True):
    """
    功能：获得期权的各个到期日期，并列出某个到期日的期权合约
    """
    import yfinance as yf
    opt = yf.Ticker(ticker)
            
    #处理称为规范日期
    from datetime import datetime
    mdate2 = datetime.strptime(mdate, '%Y-%m-%d')
    mdate3 = datetime.strftime(mdate2,'%Y-%m-%d')
    
    #获得一个到期日的所有期权合约
    try:
        optlist = opt.option_chain(mdate3)
    except:
        print("#Error(option_chain): failed to get option chain for",ticker,'\b@',mdate)
        return None,None    
    
    opt_call=optlist.calls
    opt_call['underlyingAsset']=ticker
    opt_call['optionType']='Call'
    
    inttodate=lambda x: int2date(x)
    opt_call['date']=opt_call['lastTradeDate'].apply(inttodate)
    opt_call['maturity']=mdate3
    
    collist=['contractSymbol','underlyingAsset','optionType','date', \
             'maturity','strike','lastPrice','impliedVolatility','volume', \
             'inTheMoney','currency']
    opt_call2=opt_call[collist].copy()
    num_call=len(opt_call2)
    num_call_ITM=len(opt_call2[opt_call2['inTheMoney']==True])
    num_call_OTM=num_call-num_call_ITM
    
    strike_min=min(opt_call2['strike'])
    strike_max=max(opt_call2['strike'])
    currency=opt_call2['currency'][0]
    
    opt_put=optlist.puts
    opt_put['underlyingAsset']=ticker
    opt_put['optionType']='Put'
    opt_put['date']=opt_put['lastTradeDate'].apply(inttodate)
    opt_put['maturity']=mdate3
    opt_put2=opt_put[collist].copy()
    num_put=len(opt_put2)
    num_put_ITM=len(opt_put2[opt_put2['inTheMoney']==True])
    num_put_OTM=num_put-num_put_ITM
    
    if not printout:
        return  opt_call2, opt_put2

    import datetime
    today = datetime.date.today()    
    
    print("\n===== 期权链的结构 =====")
    print("标的资产:",codetranslate(ticker))
    print("到期日期:",mdate)
    print("看涨期权:",num_call)
    print("    实值/虚值:",num_call_ITM,'/',num_call_OTM)
    print("看跌期权:",num_put)
    print("    实值/虚值:",num_put_ITM,'/',num_put_OTM)
    print("最低/最高行权价:",strike_min,'/',strike_max,currency)
    
    import datetime
    today = datetime.date.today()    
    print("*数据来源: 雅虎财经,",today)
    
    #绘图
    df1=opt_call2.copy()
    df1.sort_values(by=['strike'],axis=0,ascending=[True],inplace=True) 
    df1.set_index(['strike'],inplace=True)
    colname1='lastPrice'; label1='看涨期权价格'
    
    df2=opt_put2.copy()
    df2.sort_values(by=['strike'],axis=0,ascending=[True],inplace=True) 
    df2.set_index(['strike'],inplace=True)
    colname2='lastPrice'; label2='看跌期权价格'  
    
    ylabeltxt='期权价格('+currency+')'
    titletxt="期权价格与其标的资产行权价格的关系"
    footnote="行权价("+currency+") -->\n"+ \
        "标的资产: "+codetranslate(ticker)+ \
        ", "+"到期日: "+mdate+ \
        "\n数据来源: 雅虎财经, "+str(today)
    plot_2lines(df1,colname1,label1,df2,colname2,label2, \
                ylabeltxt,titletxt,footnote)    
    
    return  opt_call2, opt_put2

if __name__=='__main__':
    ticker='AAPL'    
    mdate='2022-6-17'    
    dfc,dfp=option_chain(ticker,mdate)

#==============================================================================
if __name__ =="__main__":
    ticker='AAPL'
    lastndays=7

#def predict_stock_trend_by_option(ticker,lastndays=7,power=4):
def predict_stock_trend_by_option(ticker,lastndays=7):    
    """
    功能：根据股票期权行权价及交易量预测股价，本函数已废弃！
    """
    
    try:
        datelist=option_maturity(ticker,printout=False)
    except:
        print("#Error(predict_stock_price_by_option): option info not found for stock",ticker)
    if datelist is None:
        print("Warning(predict_stock_price_by_option): options not found for stock",ticker)
        return None
    print("\nFound stock options with",len(datelist),"maturity dates")

    import pandas as pd
    df=pd.DataFrame(columns=['Ticker','Date','WA Strike','Total Options', \
                             'Calls in Total%','OTM in Calls%','OTM Calls in Total%', \
                             'Puts in Total%','OTM in Puts%','OTM Puts in Total%'])    
    for d in datelist:
        print("Processing options matured on",d)
        opt_call,opt_put=option_chain(ticker,d,printout=False)
        
        numofcalls=len(opt_call)
        opt_call_otm=opt_call[opt_call['inTheMoney']==False]
        numofotmcalls=len(opt_call_otm)
        
        numofputs=len(opt_put)
        opt_put_otm=opt_put[opt_put['inTheMoney']==False]
        numofotmputs=len(opt_put_otm)
        
        totalopts=numofcalls+numofputs
        
        callsintotal=round(numofcalls/totalopts*100,2)
        if numofcalls != 0:
            otmincalls=round(numofotmcalls/numofcalls*100,2)
        else:
            otmincalls=0
        otmcallsintotal=round(numofotmcalls/totalopts*100,2)
        
        putsintotal=round(numofputs/totalopts*100,2)
        if numofputs != 0:
            otminputs=round(numofotmputs/numofputs*100,2)
        else:
            otminputs=0
        otmputsintotal=round(numofotmputs/totalopts*100,2)
        
        opts=pd.concat([opt_call,opt_put])
        opts.sort_values('date',ascending=False,inplace=True)
        lasttradedate=list(opts['date'])[0]
        lastndate=date_adjust(lasttradedate, adjust=-lastndays)
        
        #取最近几天的期权交易
        opts2=opts[opts['date']>=lastndate] 
        opts2.dropna(inplace=True)
        wa_strike=round(((opts2['strike']*opts2['volume']).sum())/(opts2['volume'].sum()),2)
        
        s=pd.Series({'Ticker':ticker,'Date':d,'WA Strike':wa_strike, \
                     'Total Options':totalopts, \
                     'Calls in Total%':callsintotal,'OTM in Calls%':otmincalls,'OTM Calls in Total%':otmcallsintotal, \
                     'Puts in Total%':putsintotal,'OTM in Puts%':otminputs,'OTM Puts in Total%':otmputsintotal})
        df=df.append(s,ignore_index=True)        

    #建立日期索引
    todatetime=lambda x:pd.to_datetime(x)
    df['date']=df['Date'].apply(todatetime)
    df2=df.set_index(['date'])  
    
    #绘图1：Calls vs Puts比例
    import datetime
    today = datetime.date.today()    
    
    colname1='Calls in Total%'
    label1='Calls in Total%'
    colname2='Puts in Total%'
    label2='Puts in Total%'
    ylabeltxt='Percentage'
    titletxt="Stock Option Comparison: "+ticker+", Calls vs Puts"

    footnote="Source: Yahoo Finance, "+str(today)
    plot_line2(df2,ticker,colname1,label1, \
               df2,ticker,colname2,label2, \
               ylabeltxt,titletxt,footnote)
    
    #绘图2：OTM Calls vs OTM Puts相对比例
    colname1='OTM in Calls%'
    label1='OTM in Calls%%'
    colname2='OTM in Puts%'
    label2='OTM in Puts%'
    ylabeltxt='Percentage'
    titletxt="Stock Option Relative Proportion Comparison\n("+ticker+", OTM Calls vs OTM Puts)"

    footnote="Source: Yahoo Finance, "+str(today)
    plot_line2(df2,ticker,colname1,label1, \
               df2,ticker,colname2,label2, \
               ylabeltxt,titletxt,footnote)
    
    #绘图3：OTM Calls vs OTM Puts绝对比例
    colname1='OTM Calls in Total%'
    label1='OTM Calls in Total%%'
    colname2='OTM Puts in Total%'
    label2='OTM Puts in Total%'
    ylabeltxt='Percentage'
    titletxt="Stock Option Absolute Proportion Comparison\n("+ticker+", OTM Calls vs OTM Puts)"

    footnote="Source: Yahoo Finance, "+str(today)
    plot_line2(df2,ticker,colname1,label1, \
               df2,ticker,colname2,label2, \
               ylabeltxt,titletxt,footnote)

    #绘图4：股价预测
    df2x=df2.drop(df2[df2['Total Options']<=1].index)
    """
    colname='WA Strike'
    collabel='Predicted Stock Price'
    ylabeltxt=''
    titletxt="Predicting Stock Prices via Stock Options: "+ticker
    footnote="Source: Yahoo Finance, "+str(today)
    plot_line(df2x,colname,collabel,ylabeltxt,titletxt,footnote,power=power)
    """
    
    #打印预测结果
    compotm=lambda x:compare_otm(x)
    df2x['Trend']=df2x.apply(compotm,axis=1)  #axis=1表示按行操作
    """
    collist=['Date','Trend','WA Strike','Total Options','Calls in Total%','OTM in Calls%', \
            'Puts in Total%','OTM in Puts%']
    """
    collist=['Date','Trend','Total Options','Calls in Total%','Puts in Total%', \
             'OTM in Calls%','OTM in Puts%','OTM Calls in Total%','OTM Puts in Total%']
    df3=df2x[collist]
    df3.rename(columns={'Total Options':'TotalOptions'}, inplace = True)
    df3.rename(columns={'Calls in Total%':'Calls%','Puts in Total%':'Puts%'}, inplace = True)
    df3.rename(columns={'OTM in Calls%':'OTMinCalls%','OTM in Puts%':'OTMinPuts%'}, inplace = True)
    df3.rename(columns={'OTM Calls in Total%':'OTM Calls/Total%','OTM Puts in Total%':'OTM Puts/Total%'}, inplace = True)
    
    print("\n ======= Predicting Stock Price Trend via Stock Option Configuration: "+ticker+" =======")
    #设置打印对齐
    pd.set_option('display.max_columns', 1000)
    pd.set_option('display.width', 1000)
    pd.set_option('display.max_colwidth', 1000)
    pd.set_option('display.unicode.ambiguous_as_wide', True)
    pd.set_option('display.unicode.east_asian_width', True)
    print(df3.to_string(index=False))
    
    lastdate,lastprice=get_last_close(ticker)
    print(" Note:")
    print("   1) Recent price:",lastprice,"\b,",lastdate,'\b.')
    print("   2) +(-) predicts higher(lower) price than recent, ++(--) for more likely, +/- for undetermined trend.")
    print("   3) Option trade period: recent "+str(lastndays)+" days. No stock splits in the period. Dates with only 1 option removed.")
    #print("   4) Removed those samples with only one option, if any.")
    print(" "+footnote+'.')
    
    return df2




#================================================================
if __name__=='__main__':
    intdate=1604350755

def int2date(intdate):
    """
    功能：将数字转换为日期，10位数字为秒级，13位的为毫秒级
    输入：使用10位或13位整数表示的日期时间
    返回：yyyy-mm-dd hh:mm:ss
    """
    import time
    intstr=str(intdate)

    if len(intstr) == 10:
        tupTime = time.localtime(intdate)   #秒时间戳
        #standardTime = time.strftime("%Y-%m-%d %H:%M:%S", tupTime)
        standardTime = time.strftime("%Y-%m-%d", tupTime)
        return standardTime    

    if len(intstr)==13:
        timeNum=intdate     #毫秒时间戳
        timeTemp = float(timeNum/1000)
        tupTime = time.localtime(timeTemp)
        #standardTime = time.strftime("%Y-%m-%d %H:%M:%S", tupTime)
        standardTime = time.strftime("%Y-%m-%d", tupTime)
        return standardTime

    #直接为日期时间结构
    standardTime = intdate.strftime("%Y-%m-%d")
    return standardTime

if __name__=='__main__':
    intdate1=1604349847
    int2date(intdate1)
    
    intdate2=1566366547705
    int2date(intdate2)
    
    import pandas as pd
    intdate3=pd.to_datetime('2020-11-03')
    int2date(intdate3)
#==============================================================================
#==============================================================================
def compare_otm(row,spread=1.0):
    """
    功能：比较OTM Call vs OTM Put的绝对比例大小，赋值+/-
    输入：df的一行
    输出：+/-
    """
    if row['OTM Calls in Total%'] > row['OTM Puts in Total%']:
        result='+'
        if (row['Calls in Total%'] > row['Puts in Total%']) and ((row['OTM in Calls%'] > row['OTM in Puts%'])):
                result='++'
    
    if row['OTM Calls in Total%'] < row['OTM Puts in Total%']:
        result='-'
        if (row['Calls in Total%'] < row['Puts in Total%']) and ((row['OTM in Calls%'] < row['OTM in Puts%'])):
                result='--'

    if abs(row['OTM Calls in Total%'] - row['OTM Puts in Total%']) < spread:
        result='+/-'

    return result

#==============================================================================
def call_timevalue(row):
    """
    功能：计算看涨期权的时间价值
    输入：df的一行
    输出：期权的时间价值
    """
    #是否实值期权
    if row['inTheMoney']:
        tv=row['lastPrice']-row['lastSPrice']+row['strike']
    else:
        tv=row['lastPrice']
    return tv

def put_timevalue(row):
    """
    功能：计算看跌期权的时间价值
    输入：df的一行
    输出：期权的时间价值
    """
    #是否实值期权
    if row['inTheMoney']:
        tv=row['lastPrice']+row['lastSPrice']-row['strike']
    else:
        tv=row['lastPrice']
    return tv

#==============================================================================   

if __name__ =="__main__":
    ticker='JD'
    lastndays=2

def stock_trend_by_option(ticker,lastndays=7,cutoff=[1.1,5.0,10.0]):    
    """
    功能：根据股票期权预测股价
    算法：
    1、计算虚值看涨/看跌期权数量比例
    2、计算虚值看涨/看跌期权交易金额比例
    3、若虚值看涨期权占优，为看涨，并据此估计未来股价；
    4、若虚值看跌期权占优，为看跌，并据此估计未来股价；
    5、否则，为不明确
    返回：股票期权明细
    """
    DEBUG=True
    print("\nSearching for the option chain of stock",ticker,'...',end='')
    
    try:
        datelist=option_maturity(ticker,printout=False)
    except:
        print("\n  #Error(stock_trend_by_option): option info not found for stock",ticker)
        return None
    if datelist is None:
        print("\n  Warning(stock_trend_by_option): options not found for stock",ticker)
        return None
    print("found",len(datelist),"maturity dates")

    #最新的股价
    print("Searching for the recent close price of stock",ticker,'...',end='')
    try:
        lastsdate,lastsprice=get_last_close(ticker)
    except:
        print("\n#Error(stock_trend_by_option): failed in retrieving close price for stock",ticker)
        return None
    if (lastsdate is None) or (lastsprice is None):
        print("\n#Error(stock_trend_by_option): retrieving close price failed for stock",ticker)
        return None
    print(lastsprice,'\b,',lastsdate)

    import pandas as pd
    df=pd.DataFrame(columns=['Ticker','Date','Trend','Estimated Price', \
                             'OTM Volume Call/Put','OTM Amount Call/Put','Spot Date','Spot Price'])    
    for d in datelist:
        print("Analyzing options matured on",d,'...')
        opt_call,opt_put=option_chain(ticker,d,printout=False)
        if (len(opt_call) == 0) or (len(opt_put) == 0):
            print("  #Error(stock_trend_by_option): failed in retrieving stock option matured on",d)
            break
            return None        
        currency=list(opt_call['currency'])[0]
        
        ##########处理看涨期权##########
        if DEBUG: print("  Call volume: ",end='')
        opt_call['lastSDate']=lastsdate
        opt_call['lastSPrice']=lastsprice
        
        #去掉无交易的期权
        opt_call.dropna(inplace=True)
        #只保留过去lastndays日历日的交易
        lasttdate=max(list(opt_call['date']))
        fromtdate=date_adjust(lasttdate, adjust=-lastndays)        
        opt_call=opt_call[opt_call['date'] >= fromtdate]    
        numofcalls=sum(opt_call['volume'])
        
        """
        #计算期权的时间价值：不一致，无法进一步利用
        calltv=lambda x:call_timevalue(x)
        opt_call['timeValue']=opt_call.apply(calltv,axis=1)  #axis=1表示按行操作
        """
        
        #计算虚值期权的交易数量、交易金额和加权价格
        opt_call_otm=opt_call[opt_call['inTheMoney']==False]
        numofotmcalls=sum(opt_call_otm['volume'])
        amtofotmcalls=round(sum(opt_call_otm['lastPrice']*opt_call_otm['volume']),2)
        avopofotmcalls=round((amtofotmcalls/numofotmcalls),2)
        if DEBUG: print("total",int(numofcalls),'\b, OTM',int(numofotmcalls),end='')

        #基于加权价格判断行权价
        opt_call_otm.sort_values('lastPrice',ascending=True,inplace=True)
        strikelist=list(opt_call_otm['strike'])
        lastPricelist=list(opt_call_otm['lastPrice'])+[avopofotmcalls]
        lastPricelist=sorted(lastPricelist,reverse=False)
        avoppos=lastPricelist.index(avopofotmcalls)
        if avoppos > 0:
            op1=lastPricelist[avoppos-1]
            strike1=strikelist[avoppos-1]
            op2=lastPricelist[avoppos+1]
            strike2=strikelist[avoppos]            
            avspofotmcalls=round(min(strike1,strike2)+abs(strike1-strike2)*((avopofotmcalls-min(op1,op2))/abs(op1-op2)),2)
        else:
            avspofotmcalls=strikelist[avoppos]
        if DEBUG: print(", done!")
        
        ##########处理看跌期权##########
        if DEBUG: print("  Put volume: ",end='')
        opt_put['lastSDate']=lastsdate
        opt_put['lastSPrice']=lastsprice
        
        #去掉无交易的期权
        opt_put.dropna(inplace=True)
        #只保留过去lastndays日历日的交易
        lasttdate=max(list(opt_put['date']))
        fromtdate=date_adjust(lasttdate, adjust=-lastndays)        
        opt_put=opt_put[opt_put['date'] >= fromtdate] 
        numofputs=sum(opt_put['volume'])
        
        """        
        #计算期权的时间价值：不一致，无法进一步利用
        puttv=lambda x:put_timevalue(x)
        opt_put['timeValue']=opt_put.apply(puttv,axis=1)  #axis=1表示按行操作
        """
        
        #计算虚值期权的交易数量、交易金额和加权价格
        opt_put_otm=opt_put[opt_put['inTheMoney']==False]
        numofotmputs=sum(opt_put_otm['volume'])
        amtofotmputs=round(sum(opt_put_otm['lastPrice']*opt_put_otm['volume']),2)
        avopofotmputs=round((amtofotmputs/numofotmputs),2)
        if DEBUG: print("total",int(numofputs),'\b, OTM',int(numofotmputs),end='')

        #基于加权价格判断行权价
        opt_put_otm.sort_values('lastPrice',ascending=True,inplace=True)
        strikelist=list(opt_put_otm['strike'])
        lastPricelist=list(opt_put_otm['lastPrice'])+[avopofotmputs]
        lastPricelist=sorted(lastPricelist,reverse=False)
        avoppos=lastPricelist.index(avopofotmputs)
        if avoppos > 0:
            op1=lastPricelist[avoppos-1]
            strike1=strikelist[avoppos-1]
            op2=lastPricelist[avoppos+1]
            strike2=strikelist[avoppos]            
            avspofotmputs=round(min(strike1,strike2)+abs(strike1-strike2)*((avopofotmputs-min(op1,op2))/abs(op1-op2)),2)
        else:
            avspofotmputs=strikelist[avoppos]            
        if DEBUG: print(", done!")

        #比较虚值看涨/看跌期权的数量和交易金额
        if DEBUG: print("  Evaluating stock price trend ...",end='')
        rateqty=round(numofotmcalls/numofotmputs,2)  
        rateamt=round(amtofotmcalls/amtofotmputs,2)
        trend='+/-'
        star1=cutoff[0]; star2=cutoff[1]; star3=cutoff[2]
        #加强判断：以rateamt为主，结合rateqty
        if (rateqty > 1.0) and (rateamt > star1):
            trend='+'
        if (rateqty < 1.0) and (rateamt < 1.0/star1):
            trend='-'        
        if (rateqty > star2) and (rateamt > star2):
            trend='++'
        if (1/rateqty > star2) and (1/rateamt > star2):
            trend='--'          
        if (rateqty > star3) and (rateamt > star3):
            trend='+++'
        if (1/rateqty > star3) and (1/rateamt > star3):
            trend='---'          
        
        estsprice=lastsprice
        if (trend[0] == '+') and (trend != '+/-'):
            estsprice=avspofotmcalls
        if trend[0]=='-':
            estsprice=avspofotmputs
        if trend == '+/-':
            estsprice=(avspofotmcalls + avspofotmputs)/2.0
        
        s=pd.Series({'Ticker':ticker,'Date':d,'Trend':trend,'Estimated Price':estsprice, \
            'OTM Volume Call/Put':rateqty,'OTM Amount Call/Put':rateamt,'Spot Date':lastsdate,'Spot Price':lastsprice})
        df=df.append(s,ignore_index=True)  
        if DEBUG: print(", done!")

    #建立日期索引
    todatetime=lambda x:pd.to_datetime(x)
    df['date']=df['Date'].apply(todatetime)
    df2=df.set_index(['date'])  
    
    
    #绘图1：虚值Calls vs Puts数量比例
    import datetime
    today = datetime.date.today()    
    
    df2['Benchmark']=1.0
    colname1='OTM Volume Call/Put'
    label1='虚值看涨/看跌期权合约数量比例'
    colname2='Benchmark'
    label2='等比例线'
    ylabeltxt='比例'
    titletxt="期权链结构中的合约数量: "+codetranslate(ticker)+", 虚值看涨/看跌期权比例"

    footnote="数据来源：雅虎财经, "+str(today)
    plot_line2(df2,ticker,colname1,label1, \
               df2,ticker,colname2,label2, \
               ylabeltxt,titletxt,footnote)
    
    #绘图2：OTM Calls vs OTM Puts交易金额比例
    colname1='OTM Amount Call/Put'
    label1='虚值看涨/看跌期权交易金额比例'
    titletxt="期权链结构中的交易金额: "+codetranslate(ticker)+", 虚值看涨/看跌期权比例"

    footnote="数据来源：雅虎财经, "+str(today)
    plot_line2(df2,ticker,colname1,label1, \
               df2,ticker,colname2,label2, \
               ylabeltxt,titletxt,footnote)
    
    #绘图3：预测的股价
    df2['Benchmark']=lastsprice
    colname1='Estimated Price'
    label1='预期股价'
    colname2='Benchmark'
    label2='当前股价'
    ylabeltxt='股价('+currency+')'
    titletxt="基于期权链结构的股价走势估计: "+codetranslate(ticker)

    footnote="数据来源：雅虎财经, "+str(today)
    plot_line2(df2,ticker,colname1,label1, \
               df2,ticker,colname2,label2, \
               ylabeltxt,titletxt,footnote)
    
    #打印预测结果
    collist=['Date','Trend','Estimated Price','OTM Volume Call/Put','OTM Amount Call/Put']
    df3=df2[collist]
    
    print("\n ======= 基于期权链结构的股票走势和价格预期: "+codetranslate(ticker)+" =======")
    #设置打印对齐
    pd.set_option('display.max_columns', 1000)
    pd.set_option('display.width', 1000)
    pd.set_option('display.max_colwidth', 1000)
    pd.set_option('display.unicode.ambiguous_as_wide', True)
    pd.set_option('display.unicode.east_asian_width', True)
    df3.columns=['日期','股价走势','股价预期','虚值看涨/看跌合约数量比例','虚值看涨/看跌交易金额比例']
    print(df3.to_string(index=False))
    
    print(" 注:")
    print(" 1) 当前股价: "+currency+str(lastsprice),"\b,",lastsdate,'\b.')
    print(" 2) +(-)表示股价将比当前变高(低), +/-表示趋势不明朗.")
    print(" 3) 期权交易样本期间: 最近"+str(lastndays)+"个日历日，且期间内未发生分拆.")
    print(" 4) 股价估计可能随时变化，越远期的估计可能准确度越欠佳.")
    print(" "+footnote+'.')
    
    return df2

if __name__ =="__main__":
    df=stock_trend_by_option('AAPL',7)

#==============================================================================
#==============================================================================   
#==============================================================================   
#==============================================================================

#==============================================================================











# -*- coding: utf-8 -*-
"""
本模块功能：检验VaR(在险价值)模型计算结果的有效性
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2019年7月16日
最新修订日期：2019年8月19日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
作者邮件：wdehong2000@163.com
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""

#==============================================================================
#统一屏蔽一般性警告
import warnings; warnings.filterwarnings("ignore")        
from siat.common import *
from siat.translate import *
from siat.security_prices import *
#==============================================================================
#==============================================================================
def calc_ret(price_series,groupsize=1):
    """
    功能：计算日收益率，根据分组大小，计算组内累计收益率
    输入参数：价格序列数据框，分组大小(默认为1，即不分组)
    输出：带分组的收益率序列，数据框
    """  

    #仅用于测试
    #price_series=get_price(['AAPL'],[1],'2019-1-1','2019-8-18')       
    #groupsize=3
        
    #计算日收益率
    import pandas as pd
    #rets=pd.DataFrame(price_series['Value'].pct_change())
    rets=pd.DataFrame(price_series['Close'].pct_change())
    
    import numpy as np
    if groupsize >1:        
        group=np.arange(1,groupsize)        
        colname=rets.columns.to_list()[0]
        for i in group:
            #print(i)
            newcol="Shift"+i.astype('str')
            rets[newcol]=rets[colname].shift(i)            
    rets=rets.dropna()
    rets['Ret']=rets.apply(lambda row:np.product(1+row),axis=1)-1 

    #ret_group=pd.DataFrame(rets['Ret'])  
    ret_group=rets['Ret']        
    return ret_group        
        
        
if __name__ =="__main__":
    p=get_prices_portfolio(['AAPL','MSFT'],[1,1],'2019-1-1','2019-8-18')
    rg=calc_ret(p,5)            
        
        
#==============================================================================
def VaR_normal_standard(position,ret_series,future_days=1,alpha=0.99):
    
    """
    标准正太法VaR基本算法
    输入参数：当前持有头寸金额，收益率序列(非百分比)，未来持有时间(天)，置信度
    输出参数：VaR(金额，单位与当前头寸的金额单位相同)，负数
    """
    import numpy as np
    from scipy import stats
    
    #注意：这里z为负数
    z=stats.norm.ppf(1-alpha)
    miu_daily=np.mean(ret_series)
    #print("DEBUG: miu_daily",miu_daily)
    miu_days=np.power(miu_daily+1,future_days)-1
    sigma_daily=np.std(ret_series) 
    #print("DEBUG: sigma_daily",sigma_daily)
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    ratio=miu_days+z*sigma_days
    #print("DEBUG: ratio",ratio)
    #损失份额最多100%全损
    if np.abs(ratio) > 1.0: ratio=-1.0
    VaR_days=position*ratio
    return -abs(VaR_days)


if __name__ == "__main__":
    price=get_prices_portfolio(['AAPL'],[1],'2019-1-1','2019-8-8')
    position=price['Close'][-1]
    ret=calc_ret(price)
    VaR=VaR_normal_standard(position,ret)
#==============================================================================
def VaR_normal_modified(position,ret_series,future_days=1,alpha=0.99):
    """
    功能：VaR基本算法，修正正态法
    #输入参数：持有头寸金额，日收益率序列，未来持有日期，置信度
    #输出参数：VaR(金额)
    """
    from scipy import stats
    import numpy as np

    z=np.abs(stats.norm.ppf(1-alpha))
    S=stats.skew(ret_series)
    K=stats.kurtosis(ret_series)
    
    t1=1/6*(np.power(z,2)-1)*S
    t2=1/24*(np.power(z,3)-3*z)*K
    t3=1/36*(2*np.power(z,3)-5*z)*np.power(S,2)
    t=z+t1+t2-t3
    
    miu_daily=np.mean(ret_series)
    miu_days=np.power(miu_daily+1,future_days)-1
    sigma_daily=np.std(ret_series)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    #最多100%全损
    ratio=miu_days+t*sigma_days
    if np.abs(ratio) > 1.0: ratio=-1.0
    VaR_days=position*ratio
    return -abs(VaR_days)
#==============================================================================
def VaR_historical_1d(position,ret_series,alpha=0.99):
    """
    功能：计算VaR基本算法，历史模拟法，持有1天
    输入参数：持有头寸金额，历史日收益率序列，置信度
    输出：持有一天的VaR(金额)
    """
    import numpy as np
    n=len(ret_series)
    t=int(n*(1-alpha))
    SR=np.sort(ret_series)
    A=SR[t-1]   #SR的第一个元素的序号是0
    VaR_1d=position*A
    return -abs(VaR_1d)
#==============================================================================
def VaR_montecarlo(position,ret_series, \
                future_days=1,alpha=0.99,random=10000):
    """
    功能：计算VaR基本算法，蒙特卡洛模拟法，持有多日
    输入参数：当前头寸金额，历史日收益率序列，未来持有天数，置信度，重复模拟次数
    注：重复模拟次数越多，准确率就越高，但耗时也越多
    输出：持有多天的VaR(金额)
    """
    import numpy as np
    #取得历史日收益率的均值和标准差
    miu=np.mean(ret_series)
    sigma=np.std(ret_series)
    #生成随机数序列
    np.random.seed(12345)
    #按照历史日收益率的均值和标准差重复模拟一定次数，生成新的日收益率序列
    RR=np.random.normal(miu,sigma,random)
    #基于新的日收益率序列，使用标准正态法计算VaR
    VaR_days=VaR_normal_standard(position,RR,future_days,alpha)
    
    #最多100%全损
    if abs(VaR_days) > position: VaR_days=-position
    
    return -abs(VaR_days)
#==============================================================================
def calc_VaR_1d(position,ret_series,alpha=0.99,model="montecarlo", \
                random=10000):
    """
    功能：计算一个期间的VaR，允许指定不同的模型
    输入参数：当前头寸，各个期间的收益率序列，置信度，模型，模拟次数(仅适用于蒙特卡洛模型)
    输出：VaR金额
    """
    
    future_days=1
    #判断模型选择
    modeltype=model.lower()
    if modeltype in ['normal_standard','normal standard','ns']:
        #标准正态法
        mtype='normal_standard'
        VaR=VaR_normal_standard(position,ret_series,future_days,alpha)
    elif modeltype in ['normal_modified','normal modified','nm']:
        #修正正态法
        mtype='normal_modified'
        VaR=VaR_normal_modified(position,ret_series,future_days,alpha)
    elif modeltype in ['historical','historic','history','hist']:
        #历史模拟法
        mtype='historical'
        VaR=VaR_historical_1d(position,ret_series,alpha)
    elif modeltype in ['montecarlo','monte carlo','monte_carlo','mc']:
        #蒙特卡洛模拟法
        mtype='montecarlo'
        VaR=VaR_montecarlo(position,ret_series,future_days,alpha,random)
    else:
        print("Error #1(calc_VaR_1d): Type of model unsupported")        
        print("Variable(s):",model)
        print("Models supported: normal_standard, normal_modificed, historical, montecarlo")
        return None

    return mtype,VaR
    

if __name__ == "__main__":
    tickerlist=['AAPL','MSFT']
    sharelist=[1,2]
    start='2019-1-1'
    end='2019-6-30'     
    price=get_prices_portfolio(tickerlist,sharelist,start,end)
    
    groupsize=5
    ret=calc_ret(price,groupsize)
    
    position=price['Close'][-1]
    alpha=0.99
    type,VaR=calc_VaR_1d(position,ret,alpha,model="montecarlo")
    ratio=abs(VaR/position)
    
    threshhold_expected=round(len(ret)*(1-alpha),2)
    threshhold_actual=len(ret[ret['Ret']<-ratio])


#==============================================================================
def backtest_VaR(tickerlist,sharelist,today,future_days=1, \
                alpha=0.99,pastyears=1,model="montecarlo",random=10000):
    """
    功能：检验VaR模型的有效性，历史回溯测试
    输入参数：股票代码列表，持有股数列表，当前日期，预期持有天数，置信度，
    使用历史书据的年数，模型类型，模拟次数(仅适用于蒙特卡洛法)
    """
    
    #检查当前日期的合理性
    if not check_date(today): return
    
    #计算开始日期
    startdate=get_start_date(today,pastyears)
    import pandas as pd
    enddate=pd.to_datetime(today)
    
    #下载股价数据
    price=get_prices_portfolio(tickerlist,sharelist,startdate,enddate)
    #计算收益率，以future_days作为分组大小
    ret=calc_ret(price,future_days)
    
    #计算当前持有的头寸
    position=price['Close'][-1]
    #计算VaR
    type,VaR=calc_VaR_1d(position,ret,alpha,model)
    #计算VaR比率
    ratio=abs(VaR/position)
    
    #预期：收益率低于VaR比率的期数（天数或组数）
    threshhold_expected=round(len(ret)*(1-alpha),2)
    #实际：收益率低于VaR比率的期数（天数或组数）
    threshhold_actual=len(ret[ret<-ratio])
    alpha_actual=1-threshhold_actual/len(ret)
    
    #结果判读
    if abs(alpha_actual-alpha) < 0.001:
        result="准确"
    elif alpha_actual < alpha:
        result="低估"
    else: 
        result="高估"
    
    #打印结果
    print("\n======= VaR模型：回溯测试 =======")
    print("投资组合          :",tickerlist)
    print("成分股票配置      :",sharelist)
    print("持有日期          :",today)
    print("当前头寸          :",format(round(position,2),','))
    print("预计持有天数      :",future_days,"天")
    print("置信度水平        : ",alpha*100,"%",sep='')
    print("使用的历史样本数据:",pastyears,"年")
    print("使用的VaR模型     :",type)
    
    print("\n*** 在险价值 ***")
    print("VaR金额:",format(round(VaR,2),','))
    print("VaR比率: ",round(ratio*100,2),"%",sep='')
    
    print("\n*** 回溯测试 ***")
    print("期望的置信度水平       : ",alpha*100,"%",sep='')
    print("损失超过VaR的预期天数  :",threshhold_expected,'天')
    print("损失超过VaR的实际天数  :",threshhold_actual,'天')  
    print("实际的置信度水平       : ",round(alpha_actual*100,2),"%",sep='')
    print("模型验证的回溯测试结果 :",result)
    
    return
    
if __name__ == "__main__": 
    tickerlist=['AAPL']
    sharelist=[1]
    backtest_VaR(tickerlist,sharelist,'2019-8-8',1, \
                 pastyears=1,model="montecarlo")
    backtest_VaR(tickerlist,sharelist,'2019-8-8',1, \
                 pastyears=1,model="normal_standard")    
    backtest_VaR(tickerlist,sharelist,'2019-8-8',1, \
                 pastyears=1,model="normal_modified") 
    backtest_VaR(tickerlist,sharelist,'2019-8-8',1, \
                 pastyears=1,model="historical")     
#==============================================================================



# -*- coding: utf-8 -*-


import os; os.chdir("D:/siateng")
from siate import *

import os; os.chdir("D:/siat")
from siat import *

#==============================================================================
info=get_stock_profile("MUFG","fin_rates")

wuliu=["002352.SZ","002468.SZ","2057.HK","600233.SS","002120.SZ","603056.SS","601598.SS","603967.SS","603128.SS"]
peg=compare_snapshot(wuliu,"PEG")

tickers=["601398.SS","601988.SS","601988.SS",'601288.SS','601328.SS','601658.SS','600036.SS','000001.SZ']
isgr=compare_igr_sgr(tickers,axisamp=3)
#==============================================================================
tickers=["BABA","JD","VIPS",'PDD','AMZN','WMT','EBAY','SHOP','MELI']
isgr=compare_igr_sgr(tickers)

#==============================================================================
tickers=["0883.HK","0857.HK","0386.HK",'XOM','2222.SR','OXY','BP','RDSA.AS']
pm=compare_snapshot(tickers,'Profit Margin')
roa=compare_snapshot(tickers,'ROA')
roe=compare_snapshot(tickers,'ROE')
pe=compare_snapshot(tickers,'Trailing PE',axisamp=1.8)
#==============================================================================
tickers=["601808.SS","600583.SS","600968.SS",'600871.SS','600339.SS','601857.SS','600028.SS','0883.HK']
isgr=compare_igr_sgr(tickers)

ticker='0883.HK'
igr,sgr=calc_igr_sgr(ticker)
#==============================================================================
tat=compare_history(['AMZN','JD'],'Total Asset Turnover')
fat=compare_history(['AMZN','JD'],'Fixed Asset Turnover')
cfps_eps=compare_history(['BABA'],['Cashflow per Share','BasicEPS'])

cr=compare_history(['BABA','JD'],'Cashflow per Share')
cr=compare_history(['BABA','PDD'],'Cashflow per Share')
cr=compare_history(['BABA','VIPS'],'Cashflow per Share')

tickers=['600519.SS','000858.SZ','600779.SS','000596.SZ','603589.SS']
df=compare_dupont(tickers,fsdate='2020-12-31',scale1 = 10,scale2 = 10)


#==============================================================================
cr=compare_history(['AAPL','MSFT'],'Current Ratio')
cr=compare_history(['601808.SS','600871.SS'],'Current Ratio')
cr=compare_history(['601808.SS','0883.HK'],'Current Ratio')

cr=compare_history(['601808.SS'],['Current Ratio','Quick Ratio'])


#==============================================================================
cosl=compare_history('601808.SS',['Current Ratio','Quick Ratio'])
cosl=compare_history('601808.SS',['Debt to Asset','Debt to Equity'])
cosl=compare_history('601808.SS',['Debt to Asset','Debt to Equity'],twinx=True)
tie=compare_history(['601808.SS'],'Times Interest Earned')
itr=compare_history(['601808.SS','600871.SS'],'Inventory Turnover')
rtr=compare_history(['601808.SS','600871.SS'],'Receivable Turnover')

rtr=compare_history(['601808.SS','600871.SS'],'Total Asset Turnover')
fat=compare_history(['601808.SS','600871.SS'],'Fixed Asset Turnover')
cr=compare_history(['601808.SS','600871.SS'],'Current Ratio')
qr=compare_history(['601808.SS','600871.SS'],'Quick Ratio')
d2a=compare_history(['601808.SS','600871.SS'],'Debt to Asset')

cfps=compare_history(['601808.SS','600871.SS'],'Cashflow per Share')

dtoe=compare_snapshot(tickers,'Dbt to Asset')

tickers=["0883.HK","0857.HK","0386.HK",'XOM','2222.SR','OXY','BP','RDSA.AS']
dbi=compare_dupont(tickers,fsdate='latest',scale1 = 100,scale2 = 50)

tickers=['601808.SS',"600339.SS",'600583.SS','SLB','HAL']
dbi=compare_dupont(tickers,fsdate='latest',scale1 = 100,scale2 = 50)

igr=compare_snapshot(tickers,'IGR')
#==============================================================================
tickers=["0883.HK","0857.HK","0386.HK",'XOM','2222.SR','OXY','BP','RDSA.AS']
atr=compare_tax(tickers,graph=True)
dbi=compare_dupont(tickers,fsdate='latest',scale1 = 100,scale2 = 10)
ev2r=compare_snapshot(tickers,'EV to Revenue')
ev2ebitda=compare_snapshot(tickers,'EV to EBITDA')
price=compare_snapshot(tickers,'Current Price')

tickers2=["0883.HK","0857.HK","0386.HK",'XOM','2222.SR','BP','RDSA.AS']
fpe=compare_snapshot(tickers2,'Forward PE')
pb=compare_snapshot(tickers,'Price to Book')
roa=compare_snapshot(tickers,'ROA')
roe=compare_snapshot(tickers,'ROE')

pm=compare_snapshot(tickers,'ROE')

tickerlist=['IBM','DELL','WMT'] 
df=compare_dupont(tickerlist,fsdate='latest',scale1 = 100,scale2 = 10)    

tickerlist=['DELL','WMT'] 
df=compare_dupont(tickerlist,fsdate='latest',scale1 = 100,scale2 = 10) 
#==============================================================================
tickers_cn=['600398.SS','300005.SZ','002563.SZ','002193.SZ','002269.SZ']
tickers_hk=['2331.HK','2020.HK','1368.HK','3998.HK','2313.HK']
tickers=tickers_cn+tickers_hk
gmdf=compare_dupont(tickers,fsdate='2020-12-31',scale1 = 10,scale2 = 10) 

#==============================================================================
tickers=['AMZN','EBAY','SHOP','MELI','BABA','JD','VIPS','PDD']
dtoe=compare_snapshot(tickers,'Debt to Asset')
#==============================================================================

leadings=['JNJ','PFE','MRK','LLY','VRTX','NVS','AMGN','SNY']
tpe=compare_snapshot(leadings,'Trailing PE')

mainleadings=['JNJ','PFE','MRK','VRTX','NVS','SNY']
tests=['NBIX','REGN','PRGO']
tpe=compare_snapshot(tests + mainleadings,'Trailing PE')

#==============================================================================
leadings=['JNJ','PFE','MRK','LLY','VRTX','NVS','AMGN','SNY']
fpe=compare_snapshot(leadings,'Forward PE')

mainleadings=['JNJ','PFE','VRTX','NVS','AMGN','SNY']
tests=['NBIX','REGN','PRGO']
fpe=compare_snapshot(tests + mainleadings,'Forward PE')
#==============================================================================
cp=compare_snapshot(tests+leadings,'Current Price')
#==============================================================================
leadings=['JNJ','PFE','MRK','LLY','VRTX','NVS','AMGN','SNY']
ptob=compare_snapshot(leadings,'Price to Book')

mainleadings=['JNJ','PFE','MRK','VRTX','NVS','AMGN']
tests=['NBIX','REGN','PRGO']
ptob=compare_snapshot(tests+mainleadings,'Price to Book')
#==============================================================================
leadings=['JNJ','PFE','MRK','LLY','VRTX','NVS','AMGN','SNY']
evtoebitda=compare_snapshot(leadings,'EV to EBITDA')

mainleadings=['JNJ','PFE','VRTX','NVS','AMGN','SNY']
tests=['NBIX','REGN','PRGO']
evtoebitda=compare_snapshot(tests + mainleadings,'EV to EBITDA')
#==============================================================================
leadings=['JNJ','PFE','MRK','LLY','VRTX','NVS','AMGN','SNY']
evtoebitda=compare_snapshot(leadings,'PEG')
#==============================================================================
mainleadings=['MRK','LLY','VRTX','NVS','AMGN','SNY']
tests=['NBIX','REGN','PRGO']
peg=compare_snapshot(tests + mainleadings,'PEG')
#==============================================================================
leadings=['JNJ','PFE','MRK','LLY','VRTX','NVS','AMGN','SNY']
ptos=compare_snapshot(leadings,'TTM Price to Sales')
#==============================================================================
mainleadings=['JNJ','PFE','MRK','LLY','AMGN','SNY']
tests=['NBIX','REGN','PRGO']
ptos=compare_snapshot(tests + mainleadings,'TTM Price to Sales')
#==============================================================================
leadings=['JNJ','PFE','MRK','LLY','VRTX','NVS','AMGN','SNY']
evtorev=compare_snapshot(leadings,'EV to Revenue')
#==============================================================================
mainleadings=['JNJ','PFE','MRK','VRTX','AMGN','SNY']
tests=['NBIX','REGN','PRGO']
evtorev=compare_snapshot(tests + mainleadings,'EV to Revenue')




#==============================================================================
cfps_eps=compare_history('BABA',['Cashflow per Share','BasicEPS'])
cfps_eps=compare_history('JD',['Cashflow per Share','BasicEPS'])
cfps_eps=compare_history('PDD',['Cashflow per Share','BasicEPS'])
cfps_eps=compare_history('VIPS',['Cashflow per Share','BasicEPS'])

cfps_eps=compare_history('WMT',['Cashflow per Share','BasicEPS'])

cfps_eps=compare_history('QCOM',['Cashflow per Share','BasicEPS'])

cr=compare_history(['BABA','JD'],'Cashflow per Share')
cr=compare_history(['BABA','PDD'],'Cashflow per Share')
cr=compare_history(['BABA','VIPS'],'Cashflow per Share')

tickers=['AMZN','EBAY','SHOP','MELI']
cfps=compare_snapshot(tickers,'Cashflow per Share')
cr=compare_history(['AMZN'],['Cashflow per Share','BasicEPS'])
cr=compare_history(['EBAY'],['Cashflow per Share','BasicEPS'])


#==============================================================================
tickers=['AMZN','EBAY','SHOP','MELI','BABA','JD','VIPS','PDD']
roa=compare_snapshot(tickers,'ROA')

tat=compare_history(['AMZN','JD'],'Total Asset Turnover')
fat=compare_history(['AMZN','JD'],'Fixed Asset Turnover')
pper=compare_history(['AMZN','JD'],'PPE Residual')

cr=compare_snapshot(tickers,'Current Ratio')
qr=compare_snapshot(tickers,'Quick Ratio')
dtoe=compare_snapshot(tickers,'Debt to Equity')

pm=compare_snapshot(tickers,'Profit Margin')
gm=compare_snapshot(tickers,'Gross Margin')

gm=compare_snapshot(tickers,'EBITDA Margin')
gm=compare_snapshot(tickers,'Operating Margin')

roe=compare_snapshot(tickers,'ROE')
teps=compare_snapshot(tickers,'Trailing EPS')
rg=compare_snapshot(tickers,'Revenue Growth')

rg=compare_snapshot(tickers,'Earnings Growth')
rg=compare_snapshot(tickers,'Earnings Quarterly Growth')

#==============================================================================
tickers=['AMZN','EBAY','BABA','JD','VIPS']
roa=compare_snapshot(tickers,'ROA')
roe=compare_snapshot(tickers,'ROE')
beta=compare_snapshot(tickers,'beta')

tickers=['AMZN','EBAY','SHOP','BABA','JD','PDD','VIPS']
cr=compare_snapshot(tickers,'Current Ratio')
dtoe=compare_snapshot(tickers,'Debt to Equity')
teps=compare_snapshot(tickers,'Trailing EPS')
roe=compare_snapshot(tickers,'ROE')
hpinst=compare_snapshot(['AAPL','MSFT','BRKB'],'Held Percent Institutions')

#==============================================================================
rates=get_stock_profile('AAPL',info_type='fin_rates')

tickers=['AAPL','MSFT','WMT','FB','QCOM']
cr=compare_snapshot(tickers,'Current Ratio')
beta=compare_snapshot(tickers,'beta')
dtoe=compare_snapshot(tickers,'Debt to Equity')
dtoe=compare_snapshot(tickers,'?')
teps=compare_snapshot(tickers,'Trailing EPS')
tpe=compare_snapshot(tickers,'Trailing PE')

tickers1=['AMZN','EBAY','GRPN','BABA','JD','PDD','VIPS']
gm=compare_snapshot(tickers1,'Gross Margin')
pm=compare_snapshot(tickers1,'Profit Margin')

df1,df2=compare_history('AAPL','Current Ratio')
df1,df2=compare_history('AAPL',['Current Ratio','Quick Ratio'])
df1,df2=compare_history('AAPL',['BasicEPS','DilutedEPS'])
df1,df2=compare_history('AAPL',['Current Ratio','BasicEPS'],twinx=True)
df1,df2=compare_history('AAPL',['BasicPE','BasicEPS'])
df1,df2=compare_history('AAPL',['BasicPE','BasicEPS'],twinx=True)

df1,df2=compare_history(['AAPL','MSFT'],['BasicPE','BasicEPS'])
df1,df2=compare_history(['AAPL','MSFT'],['BasicPE','BasicEPS'],twinx=True)
df1,df2=compare_history(['AAPL','MSFT'],'BasicEPS',twinx=True)

cr=compare_history(['INTL','QCOM'],'Current Ratio',twinx=True)

cr=compare_history(['600519.SS','000002.SZ'],'Current Ratio',twinx=True)
#==============================================================================

Chinabanks = ["1398.HK","0939.HK","3988.HK","1288.HK"]
USbanks=["BAC", "TD","PNC"]
Japanbanks = ["8306.T","7182.T","8411.T"]

esg=compare_snapshot(Chinabanks+USbanks+Japanbanks,'Total ESG')
ep=compare_snapshot(Chinabanks+USbanks+Japanbanks,'Environment Score')
csr=compare_snapshot(Chinabanks+USbanks+Japanbanks,'Social Score')
emp=compare_snapshot(Chinabanks+USbanks+Japanbanks,'Employees')
gov=compare_snapshot(Chinabanks+USbanks+Japanbanks,'Governance Score')
roe=compare_snapshot(Chinabanks+USbanks+Japanbanks,'ROE')


cnnr=['2330.TW','2317.TW','2474.TW','3008.TW','2454.TW']
usnr=['SLB','COP','HAL','OXY','FCX']
otns=['5713.T','1605.T','5020.T']

esg=compare_snapshot(cnnr+usnr+otns,'Total ESG')
ep=compare_snapshot(cnnr+usnr+otns,'Environment Score')
csr=compare_snapshot(cnnr+usnr+otns,'Social Score')
gov=compare_snapshot(cnnr+usnr+otns,'Governance Score')


ep=compare_snapshot(['9988.HK','9618.HK','0700.HK'],'Environment Score')

#==============================================================================

market={'Market':('China','^HSI')}
stocks={'0700.HK':3,'9618.HK':2,'9988.HK':1}
portfolio=dict(market,**stocks)

_,_,stocklist,_=decompose_portfolio(portfolio)
collist=['symbol','totalEsg','environmentScore','socialScore','governanceScore']
sust=pd.DataFrame(columns=collist)
    for t in stocklist:
        try:
            info=stock_info(t).T
        except:
            print("#Error(): esg info not available for",t)
            continue
        if (info is None) or (len(info)==0):
            print("#Error(): failed to get esg info for",t)
            continue
        sub=info[collist]
        sust=pd.concat([sust,sub])
        















# -*- coding: utf-8 -*-
"""
@author: WANG Dehong (Peter), IBS BFSU
"""

#from __future__ import print_function
from setuptools import setup, find_packages
#import sys

setup(
    name="siat",
    version="1.3.0.6",
    author="WANG Dehong (王德宏). Business School, BFSU (北京外国语大学国际商学院)",
    author_email="wdehong2000@163.com",
    description="Securities Investment Analysis Tools (siat)",
    url = "https://pypi.org/project/siat/",
    long_description="""This plugin is for use with the author's book - 
    Security Investment Analysis, where real cases 
    can be replayed, updated and re-created in different securities, different 
    time line and different measurements. The plugin is for learning purpose only, 
    not for commercial use. The author is not responsible for the results in
    real investment activities.""",
    license="Copyright (C) WANG Dehong, 2020-2021. For education purpose only!",
    packages = find_packages(),
    install_requires=[
        'pandas_datareader',
        'yfinance',
        'tushare',
        'akshare',
        'mplfinance',
        'statsmodels',
        'yahoo_earnings_calendar',
        'yahooquery',
        'pypinyin',
        'seaborn',
        'numpy',
        'scipy',
        'pandas',
        'sklearn',
    ],            
    zip_sage=False,
    include_package_data=True, # 打包包含静态文件标识
)


# -*- coding: utf-8 -*-
"""
本模块功能：区块链的创建与查询
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2018年9月18日
最新修订日期：2020年2月7日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""

import hashlib as hasher
class Block:
    """
    定义一个区块链中的区块:
    为了保证整个区块链的完整性，每一个区块都有一个唯一的哈希值，用于自我标识。
    比如比特币，每一个区块的哈希值是由区块的索引、时间戳、数据以及前一个区块的哈希，
    经过加密后得到的。其中，数据可以选取任意值。    
    """
    def __init__(self, index, timestamp, data, previous_hash):
        self.index = index
        self.timestamp = timestamp
        self.data = data
        self.previous_hash = previous_hash
        self.hash = self.hash_block()

    def hash_block(self):
        sha = hasher.sha256()
        sha.update((str(self.index) + 
                   str(self.timestamp) + 
                   str(self.data) + 
                   str(self.previous_hash)).encode('utf-8'))
        return sha.hexdigest()


import datetime as date
def create_genesis_block():
    """
    创建一个区块链中的创世区块：即第一个区块
    这个区块的索引为0，此外，它所包含的数据以及前一个区块的哈希值都是一个任意的值。
    """ 
    # Manually construct ablock with
    # index zero and arbitrary previous hash
    return Block(0, date.datetime.now(), "RootBlock", "0")


def next_block(last_block,this_data):
    """
    增加一个区块：
    还需要一个函数来生成链上更多的区块。该函数将链上前一个区块作为参数，
    为后面的区块生成数据，并返回具有带有数据的新区块。
    当生成的新区块包含了前一个区块的哈希值，区块链的完整性就会随着每个区块的增加而增加。
    这样的操作虽然看起来有点复杂，但如果不这么做，其他人就会很容易篡改链上的数据，
    甚至把整条链都给换了。所以，链上区块的哈希值就充当了密码证明，
    确保区块一旦被添加到区块链上，就不能被替换或者删除。
    """
    this_index =last_block.index + 1
    this_timestamp =date.datetime.now()
    #this_data = "Hey! I'm block " +str(this_index)
    this_hash = last_block.hash
    return Block(this_index, this_timestamp, this_data, this_hash)


def create_block_chain(transactions):
    """
    创建一个实际的区块链例子并打印出来
    """
    # Create the blockchain and add the genesis block
    blockchain = [create_genesis_block()]
    previous_block = blockchain[0]
    # How many blocks should we add to the chain after the genesis block
    num_of_blocks_to_add= len(transactions)
    # Add blocks to the chain
    for i in range(0, num_of_blocks_to_add):
        block_to_add= next_block(previous_block,transactions[i])
        blockchain.append(block_to_add)
        previous_block = block_to_add
        # Tell everyone about it!
        print ("#{}区块已加入区块链!".format(block_to_add.index))
        print ("上家哈希值:{}".format(block_to_add.previous_hash))
        print ("区块哈希值:{}".format(block_to_add.hash))
        print ("本区块内容:{}".format(block_to_add.data))
        print ("本块时间戳:{}\n".format(block_to_add.timestamp))
    
    return blockchain

if __name__=='__main__':
    transactions=["交易1：开户","交易2：存款1万元","交易3：取款5千元"]
    blockchain=create_block_chain(transactions)


def append_block_chain(blockchain,transactions):
    """
    增加区块至现有的一个区块链
    """
    l = len(blockchain)
    previous_block = blockchain[l-1]
    # How many blocks should we add to the chain after the genesis block
    num_of_blocks_to_add= len(transactions)
    # Add blocks to the chain
    for i in range(0, num_of_blocks_to_add):
        block_to_add= next_block(previous_block,transactions[i])
        blockchain.append(block_to_add)
        previous_block = block_to_add
        # Tell everyone about it!
        print ("#{}区块已加入区块链!".format(block_to_add.index))
        print ("上家哈希值:{}".format(block_to_add.previous_hash))
        print ("区块哈希值:{}".format(block_to_add.hash))
        print ("本区块内容:{}".format(block_to_add.data))
        print ("本块时间戳:{}\n".format(block_to_add.timestamp))
    
    return blockchain

if __name__=='__main__':
    transactions2=["交易A：转账3千元","交易B：取款1千元"]
    blockchain=append_block_chain(blockchain,transactions2)


def search_block_chain(blockchain,word):
    """
    搜索一个区块链的数据内容
    """
    found=0
    # 遍历整个区块链
    for i in range(0, len(blockchain)):
        data=blockchain[i].data
        if data.find(word) != -1:
            found=found+1
            print ("找到区块#{}".format(blockchain[i].index))
            #print ("上家哈希值:{}".format(blockchain[i].previous_hash))
            #print ("区块哈希值:{}".format(blockchain[i].hash))
            print ("本区块内容:{}".format(blockchain[i].data))
            print ("本块时间戳:{}\n".format(blockchain[i].timestamp))
    if found == 0: print("抱歉，区块链中未能找到搜索关键字："+word)
    else: print("共找到"+str(found)+"个区块含有搜索关键字："+word) 
    
    return

if __name__=='__main__':
    search_block_chain(blockchain,"万元")



# -*- coding: utf-8 -*-
"""
本模块功能：上市公司的财务报表分析，数据层
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2020年9月1日
最新修订日期：2020年9月8日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！

注释：容易出现无规律的抓取失败，需要多次重复才能成功。
本部分功能作为备胎使用，当yfinance和yahooquery的财务报表功能失效时。
"""

#==============================================================================
#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
from siat.common import *
from siat.translate import *
from siat.security_prices import *
#==============================================================================
#本模块直接使用爬虫抓取雅虎财经
#==============================================================================
class YahooFin():
    
    BASE_URL = 'https://query1.finance.yahoo.com/v10/finance/quoteSummary/'

    def __init__(self, ticker):
      """Initiates the ticker
          Args: ticker (str): Stock-ticker Ex. 'AAPL'
      """
      self.ticker = ticker

    def make_request(self, url):
        """Makes a GET request"""
        import requests
        return requests.get(url)

    def get_data(self):
      """Returns a json object from a GET request"""
      try:
          return self.make_request(self.url).json()
      except KeyError as e:
            print("#Error(get_data):",e)
        
    def data(self):
      """Returns query result from json object"""
      data_temp = self.get_data()
      try:
            return data_temp.get('quoteSummary').get("result")
      except KeyError as e:
            print("Something went wrong:",e)

    def convert_timestamp(self, raw):
      """Converts UNIX-timestamp to YYYY-MM-DD"""
      from datetime import datetime
      return datetime.utcfromtimestamp(raw).strftime('%Y-%m-%d')

    def extract_raw(func):
      """Decorator to remove keys from from json data 
      that is retreived from the yahoo-module
      """
      def wrapper_extract_raw(self, *args, **kwargs):
        sheet = func(self)
        for items in sheet:
            for key, value in items.items():
                if type(value) == dict and 'fmt' in value:
                    del value['fmt']
                if type(value) == dict and 'longFmt' in value:
                    del value['longFmt']
        return sheet
      return wrapper_extract_raw

    def create_dict(self):
      """Creates a dict from extracted data"""
      balance_sheet = []
      temp_data = self._dict
      for d in temp_data:
            temp_dict = {}
            for key, value in d.items():
                if type(value) == dict and 'raw' in value:
                    v = value['raw']
                    temp_dict[key] = v
            balance_sheet.append(temp_dict)
      return balance_sheet


    def to_df(self):
      """Creates a pandas Dataframe from dict"""
      import pandas as pd
      self._df = pd.DataFrame.from_dict(self.create_dict())
      for index, row in self._df.iterrows():
            self._df.loc[index, 'endDate'] = self.convert_timestamp(self._df.at[index, 'endDate'])
      self._df = self._df.iloc[::-1]
      return self._df

#==============================================================================
class BalanceSheetQ(YahooFin):

  def __init__(self, ticker):
    super().__init__(ticker)
    self._module = 'balanceSheetHistoryQuarterly'
    self._url = (f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/'
                    f'{self.ticker}?'
                    f'modules={self.module}')
    self._dict = self._balance_sheet()
    self._df = None

  @property
  def module(self):
      return self._module

  @property
  def url(self):
      return self._url

  @property
  def df(self):
      return self._df

  @YahooFin.extract_raw
  def _balance_sheet(self):
    """Returns a balance sheet statement"""
    data = self.data()
    query =  data[0]
    balance_sheet_qty = query['balanceSheetHistoryQuarterly']
    balance_sheet_statements = balance_sheet_qty['balanceSheetStatements']
    return balance_sheet_statements

#==============================================================================
def get_bsQ(ticker):
    """
    功能： 获取最近4个季度资产负债表
    注意：数字单位为K
    """
    try:
        bsdata = BalanceSheetQ(ticker)
    except:
        bsdata = BalanceSheetQ(ticker)
    
    bsdf=bsdata.to_df()
    bsdf['ticker']=ticker
    bsdf['date']=bsdf['endDate']
    bsdf.set_index(['date'],inplace=True)
    return bsdf

if __name__ == '__main__':
    ticker='AAPL'
    bsq = get_bsQ(ticker)

#==============================================================================
class BalanceSheetA(YahooFin):

  def __init__(self, ticker):
    super().__init__(ticker)
    self._module = 'balanceSheetHistory'
    self._url = (f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/'
                    f'{self.ticker}?'
                    f'modules={self.module}')
    self._dict = self._balance_sheet()
    self._df = None

  @property
  def module(self):
      return self._module

  @property
  def url(self):
      return self._url

  @property
  def df(self):
      return self._df

  @YahooFin.extract_raw
  def _balance_sheet(self):
    """Returns a balance sheet statement"""
    data = self.data()
    
    #print(data)
    query =  data[0]
    balance_sheet_qty = query['balanceSheetHistory']
    balance_sheet_statements = balance_sheet_qty['balanceSheetStatements']
    return balance_sheet_statements

#==============================================================================
def get_bsA(ticker):
    """
    功能： 获取最近4个年度资产负债表
    注意：数字单位为K
    """
    try:
        bsdata = BalanceSheetA(ticker)
    except:
        bsdata = BalanceSheetA(ticker)
    
    bsdf=bsdata.to_df()
    bsdf['ticker']=ticker
    bsdf['date']=bsdf['endDate']
    bsdf.set_index(['date'],inplace=True)
    return bsdf

if __name__ == '__main__':
    ticker='AAPL'
    bsa = get_bsA(ticker)

#==============================================================================
class IncomeStatementQ(YahooFin):

  def __init__(self, ticker):
        super().__init__(ticker)
        self._module = 'incomeStatementHistoryQuarterly'
        self._url = (f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/'
                    f'{self.ticker}?'
                    f'modules={self.module}')
        self._dict = self._income_statement()
        self._df = None

  @property
  def module(self):
      return self._module

  @property
  def url(self):
      return self._url

  @property
  def df(self):
      return self._df

  @YahooFin.extract_raw
  def _income_statement(self):
    """Returns a income statement"""
    data = self.data()
    query =  data[0]
    income_statement_qty = query['incomeStatementHistoryQuarterly']
    income_statement_statements = income_statement_qty['incomeStatementHistory']
    return income_statement_statements


if __name__ == '__main__':
    isdata = IncomeStatementQ('AAPL')
    isdf=isdata.to_df()
    print(isdf)
    
#==============================================================================
def get_isQ(ticker):
    """
    功能： 获取最近4个季度利润表
    注意：数字单位为K
    """
    try:
        isdata = IncomeStatementQ(ticker)
    except:
        isdata = IncomeStatementQ(ticker)
    
    isdf=isdata.to_df()
    
    isdf['ticker']=ticker
    isdf['date']=isdf['endDate']
    isdf.set_index(['date'],inplace=True)
    return isdf

if __name__ == '__main__':
    ticker='AAPL'
    isq = getQ(ticker)

#==============================================================================  
class IncomeStatementA(YahooFin):

  def __init__(self, ticker):
        super().__init__(ticker)
        self._module = 'incomeStatementHistory'
        self._url = (f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/'
                    f'{self.ticker}?'
                    f'modules={self.module}')
        self._dict = self._income_statement()
        self._df = None

  @property
  def module(self):
      return self._module

  @property
  def url(self):
      return self._url

  @property
  def df(self):
      return self._df

  @YahooFin.extract_raw
  def _income_statement(self):
    """Returns a income statement"""
    data = self.data()
    #print(data)
    query =  data[0]
    income_statement_qty = query['incomeStatementHistory']
    income_statement_statements = income_statement_qty['incomeStatementHistory']
    return income_statement_statements


if __name__ == '__main__':
    isdata = IncomeStatementA('AAPL')
    isdf=isdata.to_df()
    print(isdf)

#==============================================================================
def get_isA(ticker):
    """
    功能： 获取最近4个年度利润表
    注意：数字单位为K
    """
    try:
        isdata = IncomeStatementA(ticker)
    except:
        isdata = IncomeStatementA(ticker)
    
    isdf=isdata.to_df()
    
    isdf['ticker']=ticker
    isdf['date']=isdf['endDate']
    isdf.set_index(['date'],inplace=True)
    return isdf

if __name__ == '__main__':
    ticker='AAPL'
    isa = get_isA(ticker)

#==============================================================================
class CashFlowQ(YahooFin):

  def __init__(self, ticker):
        super().__init__(ticker)
        self._module = 'cashflowStatementHistoryQuarterly'
        self._url = (f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/'
                    f'{self.ticker}?'
                    f'modules={self.module}')
        self._dict = self._cash_flow()
        self._df = None

  @property
  def module(self):
      return self._module

  @property
  def url(self):
      return self._url

  @property
  def df(self):
      return self._df

  @YahooFin.extract_raw
  def _cash_flow(self):
    """Returns a cash flow statement"""
    data = self.data()
    query =  data[0]
    #print(data)
    cash_flow_qty = query['cashflowStatementHistoryQuarterly']
    #print(cash_flow_qty)
    cash_flow_statements = cash_flow_qty['cashflowStatements']
    return cash_flow_statements


if __name__ == '__main__':
    cfdata = CashFlowQ('AAPL')
    cfdf=cfdata.to_df()
    print(cfdf)

#==============================================================================
def get_cfQ(ticker):
    """
    功能： 获取最近4个季度现金流量表
    注意：数字单位为K
    """
    try:
        cfdata = CashFlowQ(ticker)
    except:
        cfdata = CashFlowQ(ticker)
    
    cfdf=cfdata.to_df()
    cfdf['ticker']=ticker
    cfdf['date']=cfdf['endDate']
    cfdf.set_index(['date'],inplace=True)
    return cfdf

if __name__ == '__main__':
    ticker='AAPL'
    cfq = get_cfQ(ticker)


#==============================================================================
class CashFlowA(YahooFin):

  def __init__(self, ticker):
        super().__init__(ticker)
        self._module = 'cashflowStatementHistory'
        self._url = (f'https://query1.finance.yahoo.com/v10/finance/quoteSummary/'
                    f'{self.ticker}?'
                    f'modules={self.module}')
        self._dict = self._cash_flow()
        self._df = None

  @property
  def module(self):
      return self._module

  @property
  def url(self):
      return self._url

  @property
  def df(self):
      return self._df

  @YahooFin.extract_raw
  def _cash_flow(self):
    """Returns a cash flow statement"""
    data=self.data()
    #print(data)
    #print(data)
    query=data[0]
    #print(query)
    
    cash_flow_qty = query['cashflowStatementHistory']
    #print(cash_flow_qty)
    
    cash_flow_statements = cash_flow_qty['cashflowStatements']
    #print(cash_flow_qty)
    
    return cash_flow_statements


if __name__ == '__main__':
    cfdata = CashFlowA('AAPL')
    cfdf=cfdata.to_df()
    print(cfdf)

    
#==============================================================================
def get_cfA(ticker):
    """
    功能： 获取最近4个年度现金流量表
    注意：数字单位为K
    """
    try:
        cfdata = CashFlowA(ticker)
    except:
        try:
            cfdata = CashFlowA(ticker)
        except:
            cfdata = CashFlowA(ticker)
    
    cfdf=cfdata.to_df()
    cfdf['ticker']=ticker
    cfdf['date']=cfdf['endDate']
    cfdf.set_index(['date'],inplace=True)
    return cfdf

if __name__ == '__main__':
    ticker='AAPL'
    cfa = get_cfA(ticker)

#==============================================================================
def get_balance_sheet(ticker):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度资产负债表
    """
    #获取年报
    try:
        fsa = get_bsA(ticker)
    except:
        print("#Error(get_balance_sheet): no annual info available for",ticker)
        return None
    fsa['reportType']="Annual"
    colsa=list(fsa)
    
    #获取季度报
    try:
        fsq = get_bsQ(ticker)
    except:
        print("#Error(get_balance_sheet): no quarterly info available for",ticker)
        return None
    fsq['reportType']="Quarterly"    
    colsq=list(fsq)

    #两个列表中的共同部分
    cols= [x for x in colsa if x in colsq]   #两个列表中都存在
    colsdiff=[y for y in (colsa + colsq) if y not in cols] #两个列表中的不同元素
    
    #合并年度和季度报表
    import pandas as pd
    fs=pd.concat([fsa[cols],fsq[cols]])
    #排序
    fs.sort_values(by=['endDate','reportType'],inplace=True)
    #去掉重复记录
    fs.drop_duplicates(subset=['endDate'],keep='first',inplace=True)
    
    return fs    
    
if __name__ == '__main__':
    ticker='AAPL'  
    fs=get_balance_sheet('AAPL')
    fs_aapl=get_balance_sheet('AAPL')
    fs_msft=get_balance_sheet('MSFT')
    fs_c=get_balance_sheet('C')

#==============================================================================
def get_income_statements(ticker):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度利润表
    """
    #获取年报
    try:
        fsa = get_isA(ticker)
    except:
        print("#Error(get_income_statements): no annual info available for",ticker)
        return None
    fsa['reportType']="Annual"
    colsa=list(fsa)
    
    #获取季度报
    try:
        fsq = get_isQ(ticker)
    except:
        print("#Error(get_income_statements): no quarterly info available for",ticker)
        return None
    fsq['reportType']="Quarterly"    
    colsq=list(fsq)

    #两个列表中的共同部分
    cols= [x for x in colsa if x in colsq]   #两个列表中都存在
    colsdiff=[y for y in (colsa + colsq) if y not in cols] #两个列表中的不同元素
    
    #合并年度和季度报表
    import pandas as pd
    fs=pd.concat([fsa[cols],fsq[cols]])
    #排序
    fs.sort_values(by=['endDate','reportType'],inplace=True)
    #去掉重复记录
    fs.drop_duplicates(subset=['endDate'],keep='first',inplace=True)
    
    return fs    
    
if __name__ == '__main__':
    ticker='AAPL'  
    fs=get_income_statements('AAPL')
    fs_aapl=get_income_statements('AAPL')
    fs_msft=get_income_statements('MSFT')
    fs_c=get_income_statements('C')

#==============================================================================
def get_cashflow_statements(ticker):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度现金流量表
    """
    #获取年报
    try:
        fsa = get_cfA(ticker)
    except:
        print("#Error(get_cashflow_statements): no annual info available for",ticker)
        return None
    fsa['reportType']="Annual"
    colsa=list(fsa)
    
    #获取季度报
    try:
        fsq = get_cfQ(ticker)
    except:
        print("#Error(get_cashflow_statements): no quarterly info available for",ticker)
        return None
    fsq['reportType']="Quarterly"    
    colsq=list(fsq)

    #两个列表中的共同部分
    cols= [x for x in colsa if x in colsq]   #两个列表中都存在
    colsdiff=[y for y in (colsa + colsq) if y not in cols] #两个列表中的不同元素
    
    #合并年度和季度报表
    import pandas as pd
    fs=pd.concat([fsa[cols],fsq[cols]])
    #排序
    fs.sort_values(by=['endDate','reportType'],inplace=True)
    #去掉重复记录
    fs.drop_duplicates(subset=['endDate'],keep='first',inplace=True)
    
    return fs    
    
if __name__ == '__main__':
    ticker='AAPL'  
    fs=get_cashflow_statements('AAPL')
    fs_aapl=get_cashflow_statements('AAPL')
    fs_msft=get_cashflow_statements('MSFT')
    fs_c=get_cashflow_statements('C')

#==============================================================================
def get_financial_statements(ticker):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度财务报表
    """
    print("...Searching for financial info of",ticker,"\b, please wait...")
    #获取资产负债表
    try:
        fbs = get_balance_sheet(ticker)
    except:
        import time; time.sleep(3)
        try:
            fbs = get_balance_sheet(ticker)
        except:
            import time; time.sleep(5)
            try:
                fbs = get_balance_sheet(ticker)
            except:
                print("#Error(get_financial_statements): balance sheet info not available for",ticker)
                return None
    if fbs is None:
        print("#Error(get_financial_statements): balance sheet info not available for",ticker)
        return None
    
    #获取利润表
    try:
        fis = get_income_statements(ticker)
    except:
        import time; time.sleep(3)
        try:
            fis = get_income_statements(ticker)
        except:
            import time; time.sleep(5)
            try:
                fis = get_income_statements(ticker)
            except:
                print("#Error(get_financial_statements): income info not available for",ticker)
                return None
    if fis is None:
        print("#Error(get_financial_statements): income info not available for",ticker)
        return None
    
    #获取现金流量表
    try:
        fcf = get_cashflow_statements(ticker)
    except:
        import time; time.sleep(3)
        try:
            fcf = get_cashflow_statements(ticker)
        except:
            import time; time.sleep(5)
            try:
                fcf = get_cashflow_statements(ticker)
            except:
                print("#Error(get_financial_statements): cash flow info not available for",ticker)
                return None
    if fcf is None:
        print("#Error(get_financial_statements): cash flow info not available for",ticker)
        return None
    
    #合并1：资产负债表+利润表
    cols= [x for x in fis if x not in fbs]   #不在第2个列表中在第1个列表
    import pandas as pd
    fbs_fis=pd.merge(fbs,fis[cols+['endDate']],on='endDate')
    #合并2：+现金流量表
    cols= [x for x in fcf if x not in fbs_fis]   #不在第1个列表中只在第2个列表
    fbs_fis_fcf=pd.merge(fbs_fis,fcf[cols+['endDate']],on='endDate')
    
    #建立索引
    fbs_fis_fcf['date']=fbs_fis_fcf['endDate']
    bic=fbs_fis_fcf.set_index('date')
    
    #某些金融企业缺乏一些字段，特殊处理：
    cols=list(bic)
    cols_special=['inventory','totalCashFromOperatingActivities','interestExpense','dividendsPaid']
    for e in cols_special:
        if not (e in cols):
            bic[e]=0
    
    return bic    
    
if __name__ == '__main__':
    ticker='AAPL'  
    fs=get_financial_statements('AAPL')
    fs_aapl=get_financial_statements('AAPL')
    fs_msft=get_financial_statements('MSFT')
    fs_c=get_financial_statements('C')

"""
最终获得的bic表结构：
['cash',
 'shortTermInvestments',
 'netReceivables',
 'inventory',
 'otherCurrentAssets',
 'totalCurrentAssets',
 'longTermInvestments',
 'propertyPlantEquipment',
 'otherAssets',
 'totalAssets',
 'accountsPayable',
 'shortLongTermDebt',
 'otherCurrentLiab',
 'longTermDebt',
 'otherLiab',
 'totalCurrentLiabilities',
 'totalLiab',
 'commonStock',
 'retainedEarnings',
 'treasuryStock',
 'otherStockholderEquity',
 'totalStockholderEquity',
 'netTangibleAssets',
 
 'ticker',
 'reportType',
 
 'totalRevenue',
 'costOfRevenue',
 'grossProfit',
 'researchDevelopment',
 'sellingGeneralAdministrative',
 'totalOperatingExpenses',
 'operatingIncome',
 'totalOtherIncomeExpenseNet',
 'ebit',
 'interestExpense',
 'incomeBeforeTax',
 'incomeTaxExpense',
 'netIncomeFromContinuingOps',
 'netIncome',
 'netIncomeApplicableToCommonShares',
 
 'depreciation',
 'changeToNetincome',
 'changeToAccountReceivables',
 'changeToLiabilities',
 'changeToInventory',
 'changeToOperatingActivities',
 'totalCashFromOperatingActivities',
 'capitalExpenditures',
 'investments',
 'otherCashflowsFromInvestingActivities',
 'totalCashflowsFromInvestingActivities',
 'dividendsPaid',
 'netBorrowings',
 'otherCashflowsFromFinancingActivities',
 'totalCashFromFinancingActivities',
 'changeInCash',
 
 'repurchaseOfStock',
 'issuanceOfStock']
"""

#==============================================================================
def get_EPS(fsdf):
    """
    功能：抓取股票的EPS，并计算流通股数
    """
    ticker=list(fsdf['ticker'])[0]
    print("...Searching for EPS info of",ticker,"\b, please wait...")    
    
    #局限：目前取不到非美股的EPS信息！
    from yahoo_earnings_calendar import YahooEarningsCalendar
    # 实例化装饰器
    yec = YahooEarningsCalendar()
    earnings = yec.get_earnings_of(ticker)
    
    import pandas as pd
    earnings_df = pd.DataFrame.from_records(earnings) 
    if (earnings_df is None) or (len(earnings_df)==0):
        print("#Error(get_EPS): no EPS info available for",ticker)
        return False,fsdf
    
    earnings_df['datedt']=pd.to_datetime(earnings_df.startdatetime)
    datecvt=lambda x: str(x)[0:10]
    earnings_df['publish_date']=earnings_df['datedt'].apply(datecvt)
    eps=earnings_df[['ticker','epsactual','publish_date']]
    eps.dropna(inplace=True)
    if (eps is None) or (len(eps)==0):
        print("#Error(get_EPS): no EPS info available for",ticker)
        return False,fsdf
    
    eps.drop_duplicates(['publish_date'],keep='first',inplace=True)
    eps.sort_values('publish_date',inplace=True)

    epsdf=pd.DataFrame(columns=('endDate','publishDate','EPS'))

    edatelist=list(fsdf['endDate'])
    for d in edatelist:
        eps_sub=eps[eps['publish_date']>d]
        pdate=list(eps_sub['publish_date'])[0]
        epsactual=list(eps_sub['epsactual'])[0]

        #记录股价
        row=pd.Series({'endDate':d,'publishDate':pdate,'EPS':epsactual})
        epsdf=epsdf.append(row,ignore_index=True)

    fsdf1=pd.merge(fsdf,epsdf,on='endDate')
    fsdf1['outstandingStock']=fsdf1['netIncomeApplicableToCommonShares']/fsdf1['EPS']

    return True,fsdf1

if __name__ == '__main__':
    fsdf=get_financial_statements('AAPL')  
    result,fsdf1=get_EPS(fsdf)
    fsdf=get_financial_statements('0700.HK')  
    result,fsdf1=get_EPS(fsdf)

#==============================================================================
def get_PE(fsdf):
    """
    功能：计算PE
    """
    #获得各个报表的日期范围，适当扩大以规避非交易日
    start=min(list(fsdf['endDate']))
    fromdate=date_adjust(start,adjust=-30)
    end=max(list(fsdf['endDate']))
    todate=date_adjust(end,adjust=30)

    #获取股价
    ticker=list(fsdf['ticker'])[0]
    prices=get_price(ticker, fromdate, todate)
    prices['datedt']=prices.index.date
    datecvt=lambda x: str(x)[0:10]
    prices['Date']=prices['datedt'].apply(datecvt)

    #报表日期列表
    datelist_fs=list(fsdf['endDate'])
    #价格日期列表    
    datelist_price=list(prices['Date'])
    date_price_min=min(datelist_price)
    date_price_max=max(datelist_price)
    
    #股价列表
    pricelist=list(prices['Close'])
    
    import pandas as pd
    pricedf=pd.DataFrame(columns=('endDate','actualDate','Price'))
    for d in datelist_fs:
        found=False
        d1=d
        if d in datelist_price:
            found=True
            pos=datelist_price.index(d)
            p=pricelist[pos]
        else:
            while (d1 >= date_price_min) and not found:
                d1=date_adjust(d1,adjust=-1)
                if d1 in datelist_price:
                    found=True
                    pos=datelist_price.index(d1)
                    p=pricelist[pos]
            while (d1 <= date_price_max) and not found:
                d1=date_adjust(d1,adjust=1)
                if d1 in datelist_price:
                    found=True
                    pos=datelist_price.index(d1)
                    p=pricelist[pos]            
        #记录股价
        row=pd.Series({'endDate':d,'actualDate':d1,'Price':p})
        pricedf=pricedf.append(row,ignore_index=True)

    #合成表
    fsdf1=pd.merge(fsdf,pricedf,on='endDate')
    fsdf1['PE']=fsdf1['Price']/fsdf1['EPS']

    return fsdf1

if __name__ == '__main__':
    fs=get_financial_statements('AAPL')   
    fsdf=get_EPS(fs)
    fsdf1=get_PE(fsdf)

#==============================================================================
def calcFinRates(fsdf):
    """
    功能：基于财报计算各种指标
    """
    #前后填充缺失值
    fs = fsdf.fillna(method='ffill').fillna(method='bfill')
    
    #短期偿债能力指标
    #流动比率：流动资产 / 流动负债
    fs['Current Ratio']=fs['totalCurrentAssets']/fs['totalCurrentLiabilities']
    fs['Current Ratio']=round(fs['Current Ratio'],2)
    #速动比率：（流动资产-存货） / 流动负债
    fs['Quick Ratio']=(fs['totalCurrentAssets']-fs['inventory'])/fs['totalCurrentLiabilities']
    fs['Quick Ratio']=round(fs['Quick Ratio'],2)
    #现金比率: （现金+现金等价物） / 流动负债
    fs['Cash Ratio']=fs['cash']/fs['totalCurrentLiabilities']
    fs['Cash Ratio']=round(fs['Cash Ratio'],2)
    #现金流量比率：经营活动现金流量 / 流动负债
    fs['Cash Flow Ratio']=fs['totalCashFromOperatingActivities']/fs['totalCurrentLiabilities']
    fs['Cash Flow Ratio']=round(fs['Cash Flow Ratio'],2)
    #到期债务本息偿付比率：经营活动现金净流量 / （本期到期债务本金+现金利息支出）
    fs['Debt Service Ratio']=fs['totalCashFromOperatingActivities']/(fs['shortLongTermDebt']+fs['interestExpense'])
    fs['Debt Service Ratio']=round(fs['Debt Service Ratio'],2)
    
    #长期偿债能力指标
    #资产负债率：负债总额 / 资产总额
    fs['Debt to Asset Ratio']=fs['totalLiab']/fs['totalAssets']
    fs['Debt to Asset Ratio']=round(fs['Debt to Asset Ratio'],2)
    #股东权益比率：股东权益总额 / 资产总额
    fs['Equity to Asset Ratio']=fs['totalStockholderEquity']/fs['totalAssets']
    fs['Equity to Asset Ratio']=round(fs['Equity to Asset Ratio'],2)
    #权益乘数：资产总额 / 股东权益总额
    fs['Equity Multiplier']=fs['totalAssets']/fs['totalStockholderEquity']
    fs['Equity Multiplier']=round(fs['Equity Multiplier'],2)
    #负债股权比率：负债总额 / 股东权益总额
    fs['Debt to Equity Ratio']=fs['totalLiab']/fs['totalStockholderEquity']
    fs['Debt to Equity Ratio']=round(fs['Debt to Equity Ratio'],2)
    #有形净值债务率：负债总额 / （股东权益-无形资产净额）
    fs['netIntangibleAssets']=fs['totalAssets']-fs['netTangibleAssets']
    fs['Debt to Tangible Net Asset Ratio']=fs['totalAssets']/(fs['totalStockholderEquity']-fs['netIntangibleAssets'])
    fs['Debt to Tangible Net Asset Ratio']=round(fs['Debt to Tangible Net Asset Ratio'],2)
    #偿债保障比率：负债总额 / 经营活动现金净流量
    fs['Debt Service Coverage Ratio']=fs['totalLiab']/fs['totalCashFromOperatingActivities']
    fs['Debt Service Coverage Ratio']=round(fs['Debt Service Coverage Ratio'],2)
    #利息保障倍数：（税前利润+利息费用）/ 利息费用
    fs['Times Interest Earned Ratio']=fs['incomeBeforeTax']/fs['interestExpense']+1
    fs['Times Interest Earned Ratio']=round(fs['Times Interest Earned Ratio'],2)
    #现金利息保障倍数：（经营活动现金净流量+付现所得税） / 现金利息支出    
    fs['Cash Interest Coverage Ratio']=(fs['totalCashFromOperatingActivities']+fs['incomeTaxExpense'])/fs['interestExpense']
    fs['Cash Interest Coverage Ratio']=round(fs['Cash Interest Coverage Ratio'],2)
    
    #营运能力指标
    #存货周转率：销售成本 / 平均存货
    fs['avgInventory']=(fs['inventory']+fs['inventory'].shift(1))/2.0  
    fs['Inventory Turnover']=fs['costOfRevenue']/fs['avgInventory']
    fs['Inventory Turnover']=round(fs['Inventory Turnover'],2)    
    #应收账款周转率：赊销收入净额 / 平均应收账款余额    
    fs['avgReceivables']=(fs['netReceivables']+fs['netReceivables'].shift(1))/2.0  
    fs['Receivable Turnover']=fs['totalRevenue']/fs['avgReceivables']
    fs['Receivable Turnover']=round(fs['Receivable Turnover'],2)
    #流动资产周转率：销售收入 / 平均流动资产余额
    fs['avgCurrentAssets']=(fs['totalCurrentAssets']+fs['totalCurrentAssets'].shift(1))/2.0  
    fs['Current Asset Turnover']=fs['totalRevenue']/fs['avgCurrentAssets']
    fs['Current Asset Turnover']=round(fs['Current Asset Turnover'],2)
    #固定资产周转率：销售收入 / 平均固定资产净额
    fs['avgPPE']=(fs['propertyPlantEquipment']+fs['propertyPlantEquipment'].shift(1))/2.0  
    fs['Fixed Asset Turnover']=fs['totalRevenue']/fs['avgPPE']
    fs['Fixed Asset Turnover']=round(fs['Fixed Asset Turnover'],2)
    #总资产周转率：销售收入 / 平均资产总额
    fs['avgTotalAssets']=(fs['totalAssets']+fs['totalAssets'].shift(1))/2.0  
    fs['Total Asset Turnover']=fs['totalRevenue']/fs['avgTotalAssets']
    fs['Total Asset Turnover']=round(fs['Total Asset Turnover'],2)
    
    #主营业务利润率=主营业务利润/主营业务收入*100%
    fs['Operating Profit%']=fs['operatingIncome']/fs['totalRevenue']*100.0
    fs['Operating Profit%']=round(fs['Operating Profit%'],2)
    
    #发展潜力指标
    #营业收入增长率：本期营业收入增长额 / 上年同期营业收入总额
    fs['Revenue Growth%']=(fs['totalRevenue']-fs['totalRevenue'].shift(1))/(fs['totalRevenue'].shift(1))*100.0
    fs['Revenue Growth%']=round(fs['Revenue Growth%'],2)
    #资本积累率：本期所有者权益增长额 / 年初所有者权益
    fs['Capital Accumulation%']=(fs['totalStockholderEquity']-fs['totalStockholderEquity'].shift(1))/(fs['totalStockholderEquity'].shift(1))*100.0
    fs['Capital Accumulation%']=round(fs['Capital Accumulation%'],2)
    #总资产增长率：本期总资产增长额 / 年初资产总额    
    fs['Total Asset Growth%']=(fs['totalAssets']-fs['totalAssets'].shift(1))/(fs['totalAssets'].shift(1))*100.0
    fs['Total Asset Growth%']=round(fs['Total Asset Growth%'],2)
    #固定资产成新率：平均固定资产净值 / 平均固定资产原值。又称“固定资产净值率”或“有用系数”
    #fs['avgPPE']=(fs['propertyPlantEquipment']+fs['propertyPlantEquipment'].shift(1))/2.0
    fs['avgLTInvestments']=(fs['longTermInvestments']+fs['longTermInvestments'].shift(1))/2.0
    fs['Fixed Asset Net Value%']=fs['avgPPE']/fs['avgLTInvestments']*100.0
    fs['Fixed Asset Net Value%']=round(fs['Fixed Asset Net Value%'],2)
    
    #其他指标

    #盈利能力指标
    #资产报酬率（息前税后）：利润总额+利息支出 / 平均资产总额        
    #fs['avgTotalAssets']=(fs['totalAssets']+fs['totalAssets'].shift(1))/2.0  
    fs['Return on Asset']=(fs['netIncome']+fs['interestExpense'])/fs['avgTotalAssets']
    fs['Return on Asset']=round(fs['Return on Asset'],2)
    fs['ROA']=fs['Return on Asset']
    #资本回报率（Return on Invested Capital，简称ROIC）
    #ROIC=NOPLAT(息前税后经营利润)/IC(投入资本)
    #NOPLAT=EBIT×(1－T)=(营业利润+财务费用－非经常性投资损益) ×(1－所得税率)
    #IC=有息负债+净资产－超额现金－非经营性资产
    fs['Return on Invested Capital']=(fs['netIncome']+fs['interestExpense'])/fs['avgTotalAssets']
    fs['Return on Invested Capital']=round(fs['Return on Invested Capital'],2)
    fs['ROIC']=fs['Return on Invested Capital']
    #净资产报酬率：净利润 / 平均净资产    
    fs['avgTotalEquity']=(fs['totalStockholderEquity']+fs['totalStockholderEquity'].shift(1))/2.0  
    fs['Return on Net Asset']=fs['netIncome']/fs['avgTotalEquity']
    fs['Return on Net Asset']=round(fs['Return on Net Asset'],2)
    #股东权益报酬率：净利润 / 平均股东权益总额
    fs['Return on Equity']=fs['netIncome']/fs['avgTotalEquity']
    fs['Return on Equity']=round(fs['Return on Equity'],2)
    fs['ROE']=fs['Return on Equity']
    #毛利率：销售毛利 / 销售收入净额    
    fs['Gross Profit%']=fs['grossProfit']/fs['totalRevenue']*100.0
    fs['Gross Profit%']=round(fs['Gross Profit%'],2)
    #销售净利率：净利润 / 销售收入净额
    fs['Net Profit%']=fs['netIncome']/fs['totalRevenue']*100.0
    fs['Net Profit%']=round(fs['Net Profit%'],2)
    #成本费用净利率：净利润 / 成本费用总额
    fs['Net Profit on Costs%']=fs['netIncome']/fs['costOfRevenue']*100.0
    fs['Net Profit on Costs%']=round( fs['Net Profit on Costs%'],2)
    #股利发放率：每股股利 / 每股利润   
    fs['Payout Ratio%']=fs['dividendsPaid']/fs['netIncome']*100.0
    fs['Payout Ratio%']=round(fs['Payout Ratio%'],2)

    ###每股指标，受EPS可用性影响    
    #每股利润：（净利润-优先股股利） / 流通在外股数。基本EPS
    #注意：流通股股数=期初commonStock-treasuryStock,加本年增加的股数issuanceOfStock*月份占比-本年减少的股数repurchaseOfStock*月份占比
    fsdf=fs.copy()
    result,fs=get_EPS(fsdf)
    if not result: 
        ticker=list(fsdf['ticker'])[0]
        print("#Error(calcFinRates): EPS info unavailable for",ticker)
        print("#Side effect: per-share related rates are not calculateable, others are still useable.")
    else:
        fs['Earnings per Share']=fs['EPS']
        #每股现金流量：（经营活动现金净流量-优先股股利） / 流通在外股数
        fs['Cash Flow per Share']=fs['totalCashFromOperatingActivities']/fs['outstandingStock']
        fs['Cash Flow per Share']=round(fs['Cash Flow per Share'],2)
        fs['CFPS']=fs['Cash Flow per Share']
        #每股股利：（现金股利总额-优先股股利） /流通在外股数    
        fs['Dividend per Share']=fs['dividendsPaid']/fs['outstandingStock']
        fs['Dividend per Share']=round(fs['Dividend per Share'],2)
        fs['DPS']=fs['Dividend per Share']
        #每股净资产：股东权益总额 / 流通在外股数  
        fs['Net Asset per Share']=fs['totalStockholderEquity']/fs['outstandingStock']
        fs['Net Asset per Share']=round(fs['Net Asset per Share'],2)
    
        #市盈率：每股市价 / 每股利润，依赖EPS反推出的流通股数量
        fsdf=fs.copy()
        fs=get_PE(fsdf)
        fs['Price Earnings Ratio']=fs['PE']
    
    fs['date']=fs['endDate']
    fs.set_index('date',inplace=True)    
    
    return fs
    
    
if __name__ == '__main__':
    ticker='AAPL'     
    fsdf=get_financial_statements('AAPL')   
    fs=calcFinRates(fsdf)

#==============================================================================
def get_financial_rates(ticker):
    """
    功能：获得股票的财务报表和财务比率
    财务报表：资产负债表，利润表，现金流量表
    财务比率：短期还债能力，长期还债能力，营运能力，盈利能力，发展能力

短期偿债能力分析：
1、流动比率，计算公式： 流动资产 / 流动负债
2、速动比率，计算公式： （流动资产-存货） / 流动负债
3、现金比率，计算公式： （现金+现金等价物） / 流动负债
4、现金流量比率，计算公式： 经营活动现金流量 / 流动负债
5、到期债务本息偿付比率，计算公式： 经营活动现金净流量 / （本期到期债务本金+现金利息支出）

长期偿债能力分析：
1、资产负债率，计算公式： 负债总额 / 资产总额
2、股东权益比率，计算公式： 股东权益总额 / 资产总额
3、权益乘数，计算公式： 资产总额 / 股东权益总额
4、负债股权比率，计算公式： 负债总额 / 股东权益总额
5、有形净值债务率，计算公式： 负债总额 / （股东权益-无形资产净额）
6、偿债保障比率，计算公式： 负债总额 / 经营活动现金净流量
7、利息保障倍数，计算公式： （税前利润+利息费用）/ 利息费用
8、现金利息保障倍数，计算公式： （经营活动现金净流量+付现所得税） / 现金利息支出

营运分析
1、存货周转率，计算公式： 销售成本 / 平均存货
2、应收账款周转率，计算公式： 赊销收入净额 / 平均应收账款余额
3、流动资产周转率，计算公式： 销售收入 / 平均流动资产余额
4、固定资产周转率，计算公式： 销售收入 / 平均固定资产净额
5、总资产周转率，计算公式： 销售收入 / 平均资产总额

盈利分析
1、资产报酬率，计算公式： 利润总额+利息支出 / 平均资产总额
2、净资产报酬率，计算公式： 净利润 / 平均净资产
3、股东权益报酬率，计算公式： 净利润 / 平均股东权益总额
4、毛利率，计算公式： 销售毛利 / 销售收入净额
5、销售净利率，计算公式： 净利润 / 销售收入净额
6、成本费用净利率，计算公式： 净利润 / 成本费用总额
7、每股利润，计算公式： （净利润-优先股股利） / 流通在外股数
8、每股现金流量，计算公式： （经营活动现金净流量-优先股股利） / 流通在外股数
9、每股股利，计算公式： （现金股利总额-优先股股利） /流通在外股数
10、股利发放率，计算公式： 每股股利 / 每股利润
11、每股净资产，计算公式： 股东权益总额 / 流通在外股数
12、市盈率，计算公式： 每股市价 / 每股利润
13、主营业务利润率=主营业务利润/主营业务收入*100%

发展分析
1、营业增长率，计算公式： 本期营业增长额 / 上年同期营业收入总额
2、资本积累率，计算公式： 本期所有者权益增长额 / 年初所有者权益
3、总资产增长率，计算公式： 本期总资产增长额 / 年初资产总额
4、固定资产成新率，计算公式： 平均固定资产净值 / 平均固定资产原值

    返回：报表+比率    
    """
    print("\nAnalyzing financial rates of",ticker,"\b, it may take lots of time...")
    
    #抓取股票的财务报表
    try:
        fsdf=get_financial_statements(ticker)
    except:
        print("......Failed to get financial statements of",ticker,"\b!")
        print("......If the stock code",ticker,"\b is correct, please try a few minutes later.")
        return None
        
    #抓取股票的稀释后EPS，计算财务比率
    fsr=calcFinRates(fsdf)
    """
    try:
        fsr=calcFinRates(fsdf)
    except:
        print("......Failed to calculate financial rates of",ticker,"\b, please retry later!")
        return None
    """    
    #整理列名：将股票代码、截止日期、报表类型排在开头
    cols=list(fsr)
    cols.remove('endDate')
    cols.remove('ticker')
    cols.remove('reportType')
    fsr2=fsr[['ticker','endDate','reportType']+cols]
    
    return fsr2


#==============================================================================
#==============================================================================
#==============================================================================
#==============================================================================
#==============================================================================
#==============================================================================
#==============================================================================











    